# -*- coding: utf-8 -*-

import simplejson as json
import hashlib
import time
import datetime
from arrow import Arrow
from flexmock import flexmock, flexmock_teardown
from .. import EloquentTestCase, mock
from ..utils import MockModel, MockQueryBuilder, MockConnection, MockProcessor

from eloquent.query.builder import QueryBuilder
from eloquent.query.grammars import QueryGrammar
from eloquent.query.processors import QueryProcessor
from eloquent.orm.builder import Builder
from eloquent.orm.model import Model
from eloquent.exceptions.orm import ModelNotFound, MassAssignmentError
from eloquent.orm.collection import Collection
from eloquent.connections import Connection
from eloquent import DatabaseManager
from eloquent.utils import basestring


class OrmModelTestCase(EloquentTestCase):

    def tearDown(self):
        flexmock_teardown()

    def test_attributes_manipulation(self):
        model = OrmModelStub()
        model.name = 'foo'
        self.assertEqual('foo', model.name)
        del model.name
        self.assertFalse(hasattr(model, 'name'))

        # TODO: mutators

    def test_dirty_attributes(self):
        model = OrmModelStub(foo='1', bar=2, baz=3)
        model.foo = 1
        model.bar = 20
        model.baz = 30

        self.assertTrue(model.is_dirty())
        self.assertTrue(model.is_dirty('foo'))
        self.assertTrue(model.is_dirty('bar'))
        self.assertTrue(model.is_dirty('baz'))
        self.assertTrue(model.is_dirty('foo', 'bar', 'baz'))

    # TODO: test calculated attributes

    def test_new_instance_returns_instance_wit_attributes_set(self):
        model = OrmModelStub()
        instance = model.new_instance({'name': 'john'})
        self.assertIsInstance(instance, OrmModelStub)
        self.assertEqual('john', instance.name)

    def test_hydrate_creates_collection_of_models(self):
        data = [
            {'name': 'john'},
            {'name': 'jane'}
        ]
        collection = OrmModelStub.hydrate(data, 'foo_connection')

        self.assertIsInstance(collection, Collection)
        self.assertEqual(2, len(collection))
        self.assertIsInstance(collection[0], OrmModelStub)
        self.assertIsInstance(collection[1], OrmModelStub)
        self.assertEqual(collection[0].get_attributes(), collection[0].get_original())
        self.assertEqual(collection[1].get_attributes(), collection[1].get_original())
        self.assertEqual('john', collection[0].name)
        self.assertEqual('jane', collection[1].name)
        self.assertEqual('foo_connection', collection[0].get_connection_name())
        self.assertEqual('foo_connection', collection[1].get_connection_name())

    def test_hydrate_raw_makes_raw_query(self):
        model = OrmModelHydrateRawStub()
        connection = MockConnection().prepare_mock()
        connection.select.return_value = []
        model.get_connection = mock.MagicMock(return_value=connection)

        def _set_connection(name):
            model.__connection__ = name

            return model

        OrmModelHydrateRawStub.set_connection = mock.MagicMock(side_effect=_set_connection)
        collection = OrmModelHydrateRawStub.hydrate_raw('SELECT ?', ['foo'])
        self.assertEqual('hydrated', collection)
        connection.select.assert_called_once_with(
            'SELECT ?', ['foo']
        )

    def test_create_saves_new_model(self):
        model = OrmModelSaveStub.create(name='john')
        self.assertTrue(model.get_saved())
        self.assertEqual('john', model.name)

    def test_find_method_calls_query_builder_correctly(self):
        result = OrmModelFindStub.find(1)

        self.assertEqual('foo', result)

    def test_find_use_write_connection(self):
        OrmModelFindWithWriteConnectionStub.on_write_connection().find(1)

    def test_find_with_list_calls_query_builder_correctly(self):
        result = OrmModelFindManyStub.find([1, 2])

        self.assertEqual('foo', result)

    def test_destroy_method_calls_query_builder_correctly(self):
        OrmModelDestroyStub.destroy(1, 2, 3)

    def test_with_calls_query_builder_correctly(self):
        result = OrmModelWithStub.with_('foo', 'bar')
        self.assertEqual('foo', result)

    def test_update_process(self):
        query = flexmock(Builder)
        query.should_receive('where').once().with_args('id', 1)
        query.should_receive('update').once().with_args({'name': 'john'})

        model = OrmModelStub()
        model.new_query = mock.MagicMock(return_value=Builder(QueryBuilder(None, None, None)))
        model._update_timestamps = mock.MagicMock()

        # TODO: events

        model.id = 1
        model.foo = 'bar'
        model.sync_original()
        model.name = 'john'
        model.set_exists(True)
        self.assertTrue(model.save())

        model.new_query.assert_called_once_with()
        model._update_timestamps.assert_called_once_with()

    def test_update_process_does_not_override_timestamps(self):
        query = flexmock(Builder)
        query.should_receive('where').once().with_args('id', 1)
        query.should_receive('update').once().with_args({'created_at': 'foo', 'updated_at': 'bar'})

        model = OrmModelStub()
        model.new_query = mock.MagicMock(return_value=Builder(QueryBuilder(None, None, None)))
        model._update_timestamps = mock.MagicMock()

        # TODO: events

        model.id = 1
        model.sync_original()
        model.created_at = 'foo'
        model.updated_at = 'bar'
        model.set_exists(True)
        self.assertTrue(model.save())

        model.new_query.assert_called_once_with()
        self.assertTrue(model._update_timestamps.called)

    # TODO: update cancelled if updating event return false

    def test_update_process_without_timestamps(self):
        query = flexmock(Builder)
        query.should_receive('where').once().with_args('id', 1)
        query.should_receive('update').once().with_args({'name': 'john'})

        model = OrmModelStub()
        model.__timestamps__ = False
        model.new_query = mock.MagicMock(return_value=Builder(QueryBuilder(None, None, None)))
        model._update_timestamps = mock.MagicMock()

        # TODO: events

        model.id = 1
        model.sync_original()
        model.name = 'john'
        model.set_exists(True)
        self.assertTrue(model.save())

        model.new_query.assert_called_once_with()
        self.assertFalse(model._update_timestamps.called)

    def test_update_process_uses_old_primary_key(self):
        query = flexmock(Builder)
        query.should_receive('where').once().with_args('id', 1)
        query.should_receive('update').once().with_args({'id': 2, 'name': 'john'})

        model = OrmModelStub()
        model.new_query = mock.MagicMock(return_value=Builder(QueryBuilder(None, None, None)))
        model._update_timestamps = mock.MagicMock()

        # TODO: events

        model.id = 1
        model.sync_original()
        model.id = 2
        model.name = 'john'
        model.set_exists(True)
        self.assertTrue(model.save())

        model.new_query.assert_called_once_with()
        self.assertTrue(model._update_timestamps.called)

    def test_timestamps_are_returned_as_objects(self):
        model = Model()
        model.set_raw_attributes({
            'created_at': '2015-03-24',
            'updated_at': '2015-03-24'
        })

        self.assertIsInstance(model.created_at, Arrow)
        self.assertIsInstance(model.updated_at, Arrow)

    def test_timestamps_are_returned_as_objects_from_timestamps_and_datetime(self):
        model = Model()
        model.set_raw_attributes({
            'created_at': datetime.datetime.utcnow(),
            'updated_at': time.time()
        })

        self.assertIsInstance(model.created_at, Arrow)
        self.assertIsInstance(model.updated_at, Arrow)

    def test_timestamps_are_returned_as_objects_on_create(self):
        model = Model()
        model.unguard()

        timestamps = {
            'created_at': datetime.datetime.now(),
            'updated_at': datetime.datetime.now()
        }

        instance = model.new_instance(timestamps)

        self.assertIsInstance(instance.created_at, Arrow)
        self.assertIsInstance(instance.updated_at, Arrow)

        model.reguard()

    def test_timestamps_return_none_if_set_to_none(self):
        model = Model()
        model.unguard()

        timestamps = {
            'created_at': datetime.datetime.now(),
            'updated_at': datetime.datetime.now()
        }

        instance = model.new_instance(timestamps)
        instance.created_at = None

        self.assertIsNone(instance.created_at)

        model.reguard()

    def test_insert_process(self):
        query = flexmock(Builder)

        model = OrmModelStub()
        query_builder = flexmock(QueryBuilder)
        query_builder.should_receive('insert_get_id').once().with_args({'name': 'john'}, 'id').and_return(1)
        model.new_query = mock.MagicMock(return_value=Builder(QueryBuilder(None, None, None)))
        model._update_timestamps = mock.MagicMock()

        # TODO: events

        model.name = 'john'
        model.set_exists(False)
        self.assertTrue(model.save())
        self.assertEqual(1, model.id)
        self.assertTrue(model.exists)
        self.assertTrue(model._update_timestamps.called)

        model = OrmModelStub()
        query_builder.should_receive('insert').once().with_args({'name': 'john'})
        model.new_query = mock.MagicMock(return_value=Builder(QueryBuilder(None, None, None)))
        model._update_timestamps = mock.MagicMock()
        model.set_incrementing(False)

        # TODO: events

        model.name = 'john'
        model.set_exists(False)
        self.assertTrue(model.save())
        self.assertFalse(hasattr(model, 'id'))
        self.assertTrue(model.exists)
        self.assertTrue(model._update_timestamps.called)

    # TODO: insert cancelled if creating event return false

    def test_delete_properly_deletes_model(self):
        query = flexmock(Builder)
        model = OrmModelStub()
        builder = Builder(QueryBuilder(None, None, None))
        query.should_receive('where').once().with_args('id', 1).and_return(builder)
        query.should_receive('delete').once()
        model.new_query = mock.MagicMock(return_value=builder)
        model._touch_owners = mock.MagicMock()

        model.set_exists(True)
        model.id = 1
        model.delete()

        self.assertTrue(model._touch_owners.called)

    def test_push_no_relations(self):
        flexmock(Builder)
        model = flexmock(Model())
        query = flexmock(QueryBuilder(MockConnection().prepare_mock(), QueryGrammar(), QueryProcessor()))
        builder = Builder(query)
        builder.get_query().should_receive('insert_get_id').once().with_args({'name': 'john'}, 'id').and_return(1)
        model.should_receive('new_query').once().and_return(builder)
        model.should_receive('_update_timestamps').once()

        model.name = 'john'
        model.set_exists(False)

        self.assertTrue(model.push())
        self.assertEqual(1, model.id)
        self.assertTrue(model.exists)

    def test_push_empty_one_relation(self):
        flexmock(Builder)
        model = flexmock(Model())
        query = flexmock(QueryBuilder(MockConnection().prepare_mock(), QueryGrammar(), QueryProcessor()))
        builder = Builder(query)
        builder.get_query().should_receive('insert_get_id').once().with_args({'name': 'john'}, 'id').and_return(1)
        model.should_receive('new_query').once().and_return(builder)
        model.should_receive('_update_timestamps').once()

        model.name = 'john'
        model.set_exists(False)
        model.set_relation('relation_one', None)

        self.assertTrue(model.push())
        self.assertEqual(1, model.id)
        self.assertTrue(model.exists)
        self.assertIsNone(model.relation_one)

    def test_push_one_relation(self):
        flexmock(Builder)
        related1 = flexmock(Model())
        query = flexmock(QueryBuilder(MockConnection().prepare_mock(), QueryGrammar(), QueryProcessor()))
        builder = Builder(query)
        builder.get_query().should_receive('insert_get_id').once().with_args({'name': 'related1'}, 'id').and_return(2)
        related1.should_receive('new_query').once().and_return(builder)
        related1.should_receive('_update_timestamps').once()

        related1.name = 'related1'
        related1.set_exists(False)

        model = flexmock(Model())
        model.should_receive('resolve_connection').and_return(MockConnection().prepare_mock())
        query = flexmock(QueryBuilder(MockConnection().prepare_mock(), QueryGrammar(), QueryProcessor()))
        builder = Builder(query)
        builder.get_query().should_receive('insert_get_id').once().with_args({'name': 'john'}, 'id').and_return(1)
        model.should_receive('new_query').once().and_return(builder)
        model.should_receive('_update_timestamps').once()

        model.name = 'john'
        model.set_exists(False)
        model.set_relation('relation_one', related1)

        self.assertTrue(model.push())
        self.assertEqual(1, model.id)
        self.assertTrue(model.exists)
        self.assertEqual(2, model.relation_one.id)
        self.assertTrue(model.relation_one.exists)
        self.assertEqual(2, related1.id)
        self.assertTrue(related1.exists)

    def test_push_empty_many_relation(self):
        flexmock(Builder)
        model = flexmock(Model())
        query = flexmock(QueryBuilder(MockConnection().prepare_mock(), QueryGrammar(), QueryProcessor()))
        builder = Builder(query)
        builder.get_query().should_receive('insert_get_id').once().with_args({'name': 'john'}, 'id').and_return(1)
        model.should_receive('new_query').once().and_return(builder)
        model.should_receive('_update_timestamps').once()

        model.name = 'john'
        model.set_exists(False)
        model.set_relation('relation_many', Collection([]))

        self.assertTrue(model.push())
        self.assertEqual(1, model.id)
        self.assertTrue(model.exists)
        self.assertEqual(0, len(model.relation_many))

    def test_push_many_relation(self):
        flexmock(Builder)
        related1 = flexmock(Model())
        query = flexmock(QueryBuilder(MockConnection().prepare_mock(), QueryGrammar(), QueryProcessor()))
        builder = Builder(query)
        builder.get_query().should_receive('insert_get_id').once().with_args({'name': 'related1'}, 'id').and_return(2)
        related1.should_receive('new_query').once().and_return(builder)
        related1.should_receive('_update_timestamps').once()

        related1.name = 'related1'
        related1.set_exists(False)

        flexmock(Builder)
        related2 = flexmock(Model())
        query = flexmock(QueryBuilder(MockConnection().prepare_mock(), QueryGrammar(), QueryProcessor()))
        builder = Builder(query)
        builder.get_query().should_receive('insert_get_id').once().with_args({'name': 'related2'}, 'id').and_return(3)
        related2.should_receive('new_query').once().and_return(builder)
        related2.should_receive('_update_timestamps').once()

        related2.name = 'related2'
        related2.set_exists(False)

        model = flexmock(Model())
        model.should_receive('resolve_connection').and_return(MockConnection().prepare_mock())
        query = flexmock(QueryBuilder(MockConnection().prepare_mock(), QueryGrammar(), QueryProcessor()))
        builder = Builder(query)
        builder.get_query().should_receive('insert_get_id').once().with_args({'name': 'john'}, 'id').and_return(1)
        model.should_receive('new_query').once().and_return(builder)
        model.should_receive('_update_timestamps').once()

        model.name = 'john'
        model.set_exists(False)
        model.set_relation('relation_many', Collection([related1, related2]))

        self.assertTrue(model.push())
        self.assertEqual(1, model.id)
        self.assertTrue(model.exists)
        self.assertEqual(2, len(model.relation_many))
        self.assertEqual([2, 3], model.relation_many.lists('id'))

    def test_new_query_returns_eloquent_query_builder(self):
        conn = flexmock(Connection)
        grammar = flexmock(QueryGrammar)
        processor = flexmock(QueryProcessor)
        conn.should_receive('get_query_grammar').and_return(grammar)
        conn.should_receive('get_post_processor').and_return(processor)
        resolver = flexmock(DatabaseManager)
        resolver.should_receive('connection').and_return(Connection(None))
        OrmModelStub.set_connection_resolver(DatabaseManager({}))

        model = OrmModelStub()
        builder = model.new_query()
        self.assertIsInstance(builder, Builder)

    def test_get_and_set_table(self):
        model = OrmModelStub()
        self.assertEqual('stub', model.get_table())
        model.set_table('foo')
        self.assertEqual('foo', model.get_table())

    def test_get_key_returns_primary_key_value(self):
        model = OrmModelStub()
        model.id = 1
        self.assertEqual(1, model.get_key())
        self.assertEqual('id', model.get_key_name())

    def test_connection_management(self):
        resolver = flexmock(DatabaseManager)
        resolver.should_receive('connection').once().with_args('foo').and_return('bar')

        OrmModelStub.set_connection_resolver(DatabaseManager({}))
        model = OrmModelStub()
        model.set_connection('foo')

        self.assertEqual('bar', model.get_connection())

    def test_to_dict(self):
        model = OrmModelStub()
        model.name = 'foo'
        model.age = None
        model.password = 'password1'
        model.set_hidden(['password'])

        # TODO: relations

        d = model.to_dict()

        self.assertIsInstance(d, dict)
        self.assertEqual('foo', d['name'])
        self.assertIsNone(d['age'])

        # TODO: relations

    def test_to_dict_includes_default_formatted_timestamps(self):
        model = Model()
        model.set_raw_attributes({
            'created_at': '2015-03-24',
            'updated_at': '2015-03-25'
        })

        d = model.to_dict()

        self.assertEqual('2015-03-24T00:00:00+00:00', d['created_at'])
        self.assertEqual('2015-03-25T00:00:00+00:00', d['updated_at'])

    def test_to_dict_includes_custom_formatted_timestamps(self):
        class Stub(Model):

            def get_date_format(self):
                return 'DD-MM-YY'

        model = Stub()
        model.set_raw_attributes({
            'created_at': '2015-03-24',
            'updated_at': '2015-03-25'
        })

        d = model.to_dict()

        self.assertEqual('24-03-15', d['created_at'])
        self.assertEqual('25-03-15', d['updated_at'])

    def test_visible_creates_dict_whitelist(self):
        model = OrmModelStub()
        model.set_visible(['name'])
        model.name = 'John'
        model.age = 28
        d = model.to_dict()

        self.assertEqual({'name': 'John'}, d)

    # TODO: hidden also hides relationship

    # TODO: to_dict uses mutators

    def test_hidden_are_ignored_when_visible(self):
        model = OrmModelStub(name='john', age=28, id='foo')
        model.set_visible(['name', 'id'])
        model.set_hidden(['name', 'age'])
        d = model.to_dict()

        self.assertIn('name', d)
        self.assertIn('id', d)
        self.assertNotIn('age', d)

    def test_fillable(self):
        model = OrmModelStub()
        model.fillable(['name', 'age'])
        model.fill(name='foo', age=28)
        self.assertEqual('foo', model.name)
        self.assertEqual(28, model.age)

    def test_unguard_allows_anything(self):
        model = OrmModelStub()
        model.unguard()
        model.guard(['*'])
        model.fill(name='foo', age=28)
        self.assertEqual('foo', model.name)
        self.assertEqual(28, model.age)
        model.reguard()

    def test_underscore_properties_are_not_filled(self):
        model = OrmModelStub()
        model.fill(_foo='bar')
        self.assertEqual({}, model.get_attributes())

    def test_guarded(self):
        model = OrmModelStub()
        model.guard(['name', 'age'])
        model.fill(name='foo', age='bar', foo='bar')
        self.assertFalse(hasattr(model, 'name'))
        self.assertFalse(hasattr(model, 'age'))
        self.assertEqual('bar', model.foo)

    def test_fillable_overrides_guarded(self):
        model = OrmModelStub()
        model.guard(['name', 'age'])
        model.fillable(['age', 'foo'])
        model.fill(name='foo', age='bar', foo='bar')
        self.assertFalse(hasattr(model, 'name'))
        self.assertEqual('bar', model.age)
        self.assertEqual('bar', model.foo)

    def test_global_guarded(self):
        model = OrmModelStub()
        model.guard(['*'])
        self.assertRaises(
            MassAssignmentError,
            model.fill,
            name='foo', age='bar', foo='bar'
        )

    # TODO: test relations

    def test_models_assumes_their_name(self):
        model = OrmModelNoTableStub()

        self.assertEqual('orm_model_no_table_stubs', model.get_table())

    # TODO: mutators cache

    def test_clone_model_makes_a_fresh_copy(self):
        model = OrmModelStub()
        model.id = 1
        model.set_exists(True)
        model.first = 'john'
        model.last = 'doe'
        model.created_at = model.fresh_timestamp()
        model.updated_at = model.fresh_timestamp()
        # TODO: relation

        clone = model.replicate()

        self.assertFalse(hasattr(clone, 'id'))
        self.assertFalse(clone.exists)
        self.assertEqual('john', clone.first)
        self.assertEqual('doe', clone.last)
        self.assertFalse(hasattr(clone, 'created_at'))
        self.assertFalse(hasattr(clone, 'updated_at'))
        # TODO: relation

        clone.first = 'jane'

        self.assertEqual('john', model.first)
        self.assertEqual('jane', clone.first)

    def test_get_attribute_raise_attribute_error(self):
        model = OrmModelStub()

        try:
            relation = model.incorrect_relation
            self.fail('AttributeError not raised')
        except AttributeError:
            pass

    def test_increment(self):
        query = flexmock()
        model_mock = flexmock(OrmModelStub, new_query=lambda: query)
        model = OrmModelStub()
        model.set_exists(True)
        model.id = 1
        model.sync_original_attribute('id')
        model.foo = 2

        model_mock.should_receive('new_query').and_return(query)
        query.should_receive('where').and_return(query)
        query.should_receive('increment')

        model.public_increment('foo')

        self.assertEqual(3, model.foo)
        self.assertFalse(model.is_dirty())

    # TODO: relationship touch_owners is propagated

    # TODO: relationship touch_owners is not propagated if no relationship result

    def test_timestamps_are_not_update_with_timestamps_false_save_option(self):
        query = flexmock(Builder)
        query.should_receive('where').once().with_args('id', 1)
        query.should_receive('update').once().with_args({'name': 'john'})

        model = OrmModelStub()
        model.new_query = mock.MagicMock(return_value=Builder(QueryBuilder(None, None, None)))

        model.id = 1
        model.sync_original()
        model.name = 'john'
        model.set_exists(True)
        self.assertTrue(model.save({'timestamps': False}))
        self.assertFalse(hasattr(model, 'updated_at'))

        model.new_query.assert_called_once_with()

    def test_casts(self):
        model = OrmModelCastingStub()
        model.first = '3'
        model.second = '4.0'
        model.third = 2.5
        model.fourth = 1
        model.fifth = 0
        model.sixth = {'foo': 'bar'}
        model.seventh = ['foo', 'bar']
        model.eighth = {'foo': 'bar'}

        self.assertIsInstance(model.first, int)
        self.assertIsInstance(model.second, float)
        self.assertIsInstance(model.third, basestring)
        self.assertIsInstance(model.fourth, bool)
        self.assertIsInstance(model.fifth, bool)
        self.assertIsInstance(model.sixth, dict)
        self.assertIsInstance(model.seventh, list)
        self.assertIsInstance(model.eighth, dict)
        self.assertTrue(model.fourth)
        self.assertFalse(model.fifth)
        self.assertEqual({'foo': 'bar'}, model.sixth)
        self.assertEqual({'foo': 'bar'}, model.eighth)
        self.assertEqual(['foo', 'bar'], model.seventh)
        
        d = model.to_dict()
        
        self.assertIsInstance(d['first'], int)
        self.assertIsInstance(d['second'], float)
        self.assertIsInstance(d['third'], basestring)
        self.assertIsInstance(d['fourth'], bool)
        self.assertIsInstance(d['fifth'], bool)
        self.assertIsInstance(d['sixth'], dict)
        self.assertIsInstance(d['seventh'], list)
        self.assertIsInstance(d['eighth'], dict)
        self.assertTrue(d['fourth'])
        self.assertFalse(d['fifth'])
        self.assertEqual({'foo': 'bar'}, d['sixth'])
        self.assertEqual({'foo': 'bar'}, d['eighth'])
        self.assertEqual(['foo', 'bar'], d['seventh'])

    def test_casts_preserve_null(self):
        model = OrmModelCastingStub()
        model.first = None
        model.second = None
        model.third = None
        model.fourth = None
        model.fifth = None
        model.sixth = None
        model.seventh = None
        model.eighth = None

        self.assertIsNone(model.first)
        self.assertIsNone(model.second)
        self.assertIsNone(model.third)
        self.assertIsNone(model.fourth)
        self.assertIsNone(model.fifth)
        self.assertIsNone(model.sixth)
        self.assertIsNone(model.seventh)
        self.assertIsNone(model.eighth)
        
        d = model.to_dict()
        
        self.assertIsNone(d['first'])
        self.assertIsNone(d['second'])
        self.assertIsNone(d['third'])
        self.assertIsNone(d['fourth'])
        self.assertIsNone(d['fifth'])
        self.assertIsNone(d['sixth'])
        self.assertIsNone(d['seventh'])
        self.assertIsNone(d['eighth'])


class OrmModelStub(Model):

    __table__ = 'stub'

    __guarded__ = []

    def get_list_items_attribute(self, value):
        return json.loads(value)

    def set_list_items_attribute(self, value):
        self._attributes['list_items'] = json.dumps(value)

    def get_password_attribute(self, _):
        return '******'

    def set_password_attribute(self, value):
        self._attributes['password_hash'] = hashlib.md5(value).hexdigest()

    def public_increment(self, column, amount=1):
        return self._increment(column, amount)

    def get_dates(self):
        return []


class OrmModelHydrateRawStub(Model):

    @classmethod
    def hydrate(cls, items, connection=None):
        return 'hydrated'


class OrmModelWithStub(Model):

    def new_query(self):
        mock = flexmock(Builder(None))
        mock.should_receive('with_').once().with_args('foo', 'bar').and_return('foo')

        return mock


class OrmModelSaveStub(Model):

    __table__ = 'save_stub'

    __guarded__ = []

    __saved = False

    def save(self, options=None):
        self.__saved = True

    def set_incrementing(self, value):
        self.__incrementing__ = value

    def get_saved(self):
        return self.__saved


class OrmModelFindStub(Model):

    def new_query(self):
        flexmock(Builder).should_receive('find').once().with_args(1, ['*']).and_return('foo')

        return Builder(None)


class OrmModelFindWithWriteConnectionStub(Model):

    def new_query(self):
        mock = flexmock(Builder)
        mock_query = flexmock(QueryBuilder)
        mock_query.should_receive('use_write_connection').once().and_return(flexmock)
        mock.should_receive('find').once().with_args(1).and_return('foo')

        return Builder(QueryBuilder(None, None, None))


class OrmModelFindManyStub(Model):

    def new_query(self):
        mock = flexmock(Builder)
        mock.should_receive('find').once().with_args([1, 2], ['*']).and_return('foo')

        return Builder(QueryBuilder(None, None, None))


class OrmModelDestroyStub(Model):

    def new_query(self):
        mock = flexmock(Builder)
        model = flexmock()
        mock_query = flexmock(QueryBuilder)
        mock_query.should_receive('where_in').once().with_args('id', [1, 2, 3]).and_return(flexmock)
        mock.should_receive('get').once().and_return([model])
        model.should_receive('delete').once()

        return Builder(QueryBuilder(None, None, None))


class OrmModelNoTableStub(Model):

    pass


class OrmModelCastingStub(Model):

    __casts__ = {
        'first': 'int',
        'second': 'float',
        'third': 'str',
        'fourth': 'bool',
        'fifth': 'boolean',
        'sixth': 'dict',
        'seventh': 'list',
        'eighth': 'json'
    }



# -*- coding: utf-8 -*-

from . import mock
from eloquent.connections.connection_interface import ConnectionInterface
from eloquent.query.processors.processor import QueryProcessor
from eloquent.database_manager import DatabaseManager
from eloquent.connectors.connection_factory import ConnectionFactory
from eloquent.query.builder import QueryBuilder
from eloquent.orm.model import Model


class MockConnection(ConnectionInterface):

    def set_reconnector(self, reconnector):
        return mock.MagicMock()

    def prepare_mock(self):
        self.table = mock.MagicMock()
        self.select = mock.MagicMock()
        self.insert = mock.MagicMock()
        self.update = mock.MagicMock()
        self.delete = mock.MagicMock()
        self.statement = mock.MagicMock()

        return self


class MockProcessor(QueryProcessor):

    def prepare_mock(self):
        self.process_select = mock.MagicMock()
        self.process_insert_get_id = mock.MagicMock()

        return self


class MockManager(DatabaseManager):

    def prepare_mock(self):
        self._make_connection = mock.MagicMock(
            side_effect=lambda name: MockConnection().prepare_mock()
        )

        return self


class MockFactory(ConnectionFactory):

    def prepare_mock(self):
        self.make = mock.MagicMock(return_value=MockConnection().prepare_mock())

        return self


class MockQueryBuilder(QueryBuilder):

    def prepare_mock(self):
        self.from__ = 'foo_table'

        return self


class MockModel(Model):

    def prepare_mock(self):
        self.get_key_name = mock.MagicMock(return_value='foo')
        self.get_table = mock.MagicMock(return_value='foo_table')
        self.get_qualified_key_name = mock.MagicMock(return_value='foo_table.foo')

        return self



# -*- coding: utf-8 -*-

import sys
import os
from unittest import TestCase
from eloquent.database_manager import DatabaseManager
from .orm.models import Model, User

PY2 = sys.version_info[0] == 2

if PY2:
    import mock
else:
    import unittest.mock as mock


class EloquentTestCase(TestCase):

    def tearDown(self):
        if hasattr(self, 'local_database'):
            os.remove(self.local_database)

    def init_database(self):
        self.local_database = '/tmp/eloquent_test_database.db'

        if os.path.exists(self.local_database):
            os.remove(self.local_database)

        self.manager = DatabaseManager({
            'default': 'sqlite',
            'sqlite': {
                'driver': 'sqlite',
                'database': self.local_database
            }
        })

        with self.manager.transaction():
            try:
                self.manager.statement(
                    'CREATE TABLE `users` ('
                    'id INTEGER PRIMARY KEY NOT NULL, '
                    'name CHAR(50) NOT NULL, '
                    'created_at DATETIME DEFAULT CURRENT_TIMESTAMP, '
                    'updated_at DATETIME DEFAULT CURRENT_TIMESTAMP'
                    ')'
                )
            except Exception:
                pass

        Model.set_connection_resolver(self.manager)

        self.manager.disconnect()

    def assertRegex(self, *args, **kwargs):
        if PY2:
            return self.assertRegexpMatches(*args, **kwargs)
        else:
            return super(EloquentTestCase, self).assertRegex(*args, **kwargs)

    def assertNotRegex(self, *args, **kwargs):
        if PY2:
            return self.assertNotRegexpMatches(*args, **kwargs)
        else:
            return super(EloquentTestCase, self).assertNotRegex(*args, **kwargs)



# -*- coding: utf-8 -*-

from ...query.expression import QueryExpression
from .relation import Relation


class HasManyThrough(Relation):

    def __init__(self, query, far_parent, parent, first_key, second_key):
        """
        :param query: A Builder instance
        :type query: Builder

        :param far_parent: The far parent model
        :type far_parent: Model

        :param parent: The parent model
        :type parent: Model

        :type first_key: str
        :type second_key: str
        """
        self._first_key = first_key
        self._second_key = second_key
        self._far_parent = far_parent

        super(HasManyThrough, self).__init__(query, parent)

    def add_constraints(self):
        """
        Set the base constraints on the relation query.

        :rtype: None
        """
        parent_table = self._parent.get_table()

        self._set_join()

        if self._constraints:
            self._query.where('%s.%s' % (parent_table, self._first_key), '=', self._far_parent.get_key())

    def get_relation_count_query(self, query, parent):
        """
        Add the constraints for a relationship count query.

        :type query: Builder
        :type parent: Builder

        :rtype: Builder
        """
        parent_table = self._parent.get_table()

        self._set_join(query)

        query.select(QueryExpression('COUNT(*)'))

        key = self.wrap('%s.%s' % (parent_table, self._first_key))

        return query.where(self.get_has_compare_key(), '=', QueryExpression(key))

    def _set_join(self, query=None):
        """
        Set the join clause for the query.
        """
        if not query:
            query = self._query

        foreign_key = '%s.%s' % (self._related.get_table(), self._second_key)

        query.join(self._parent.get_table(), self.get_qualified_parent_key_name(), '=', foreign_key)

    def add_eager_constraints(self, models):
        """
        Set the constraints for an eager load of the relation.

        :type models: list
        """
        table = self._parent.get_table()

        self._query.where_in('%s.%s' % (table, self._first_key), self.get_keys(models))

    def init_relation(self, models, relation):
        """
        Initialize the relation on a set of models.

        :type models: list
        :type relation:  str
        """
        for model in models:
            model.set_relation(relation, self._related.new_collection())

        return models

    def match(self, models, results, relation):
        """
        Match the eagerly loaded results to their parents.

        :type models: list
        :type results: Collection
        :type relation:  str
        """
        dictionary = self._build_dictionary(results)

        for model in models:
            key = model.get_key()

            if key in dictionary:
                value = self._related.new_collection(dictionary[key])

                model.set_relation(relation, value)

        return models

    def _build_dictionary(self, results):
        """
        Build model dictionary keyed by the relation's foreign key.

        :param results: The results
        :type results: Collection

        :rtype: dict
        """
        foreign = self._first_key

        dictionary = {}

        for result in results:
            key = getattr(result, foreign)
            if key not in dictionary:
                dictionary[key] = []

            dictionary[key].append(result)

        return dictionary

    def get_results(self):
        """
        Get the results of the relationship.
        """
        return self.get()

    def get(self, columns=None):
        """
        Execute the query as a "select" statement.

        :type columns: list

        :rtype: eloquent.Collection
        """
        if columns is None:
            columns = ['*']

        select = self._get_select_columns(columns)

        models = self._query.add_select(select).get_models()

        if len(models) > 0:
            models = self._query.eager_load_relations(models)

        return self._related.new_collection(models)

    def _get_select_columns(self, columns=None):
        """
        Set the select clause for the relation query.

        :param columns: The columns
        :type columns: list

        :rtype: list
        """
        if columns == ['*'] or columns is None:
            columns = ['%s.*' % self._related.get_table()]

        return columns + ['%s.%s' % (self._parent.get_table(), self._first_key)]

    def get_has_compare_key(self):
        return self._far_parent.get_qualified_key_name()



# -*- coding: utf-8 -*-

from .grammar import QueryGrammar
from ...utils import basestring


class PostgresQueryGrammar(QueryGrammar):

    _operators = [
        '=', '<', '>', '<=', '>=', '<>', '!=',
        'like', 'not like', 'between', 'ilike',
        '&', '|', '#', '<<', '>>'
    ]

    def _compile_lock(self, query, value):
        """
        Compile the lock into SQL

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :param value: The lock value
        :type value: bool or str

        :return: The compiled lock
        :rtype: str
        """
        if isinstance(value, basestring):
            return value

        if value:
            return 'FOR UPDATE'

        return 'FOR SHARE'

    def compile_update(self, query, values):
        """
        Compile an update statement into SQL

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :param values: The update values
        :type values: dict

        :return: The compiled update
        :rtype: str
        """
        table = self.wrap_table(query.from__)

        columns = self._compile_update_columns(values)

        from_ = self._compile_update_from(query)

        where = self._compile_update_wheres(query)

        return ('UPDATE %s SET %s%s %s' % (table, columns, from_, where)).strip()

    def _compile_update_columns(self, values):
        """
        Compile the columns for the update statement

        :param values: The columns
        :type values: dict

        :return: The compiled columns
        :rtype: str
        """
        columns = []

        for key, value in values.items():
            columns.append('%s = %s' % (self.wrap(key), self.parameter(value)))

        return ', '.join(columns)

    def _compile_update_from(self, query):
        """
        Compile the "from" clause for an update with a join.

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :return: The compiled sql
        :rtype: str
        """
        if not query.joins:
            return ''

        froms = []

        for join in query.joins:
            froms.append(self.wrap_table(join.table))

        if len(froms):
            return ' FROM %s' % ', '.join(froms)

        return ''

    def _compile_update_wheres(self, query):
        """
        Compile the additional where clauses for updates with joins.

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :return: The compiled sql
        :rtype: str
        """
        base_where = self._compile_wheres(query)

        if not query.joins:
            return base_where

        join_where = self._compile_update_join_wheres(query)

        if not base_where.strip():
            return 'WHERE %s' % self._remove_leading_boolean(join_where)

        return '%s %s' % (base_where, join_where)

    def _compile_update_join_wheres(self, query):
        """
        Compile the "join" clauses for an update.

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :return: The compiled sql
        :rtype: str
        """
        join_wheres = []

        for join in query.joins:
            for clause in join.clauses:
                join_wheres.append(self._compile_join_constraints(clause))

        return ' '.join(join_wheres)

    def compile_insert_get_id(self, query, values, sequence=None):
        """
        Compile an insert and get ID statement into SQL.

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :param values: The values to insert
        :type values: dict

        :param sequence: The id sequence
        :type sequence: str

        :return: The compiled statement
        :rtype: str
        """
        if sequence is None:
            sequence = 'id'

        return '%s RETURNING %s'\
               % (self.compile_insert(query, values), self.wrap(sequence))

    def compile_truncate(self, query):
        """
        Compile a truncate table statement into SQL.

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :return: The compiled statement
        :rtype: str
        """
        return {
            'TRUNCATE %s RESTART IDENTITY' % self.wrap_table(query.from__): {}
        }

    def get_marker(self):
        return '%s'



# -*- coding: utf-8 -*-

from ..exceptions.orm import ModelNotFound
from ..utils import Null, basestring
from ..query.expression import QueryExpression


class Builder(object):

    _passthru = [
        'to_sql', 'lists', 'insert', 'insert_get_id', 'pluck', 'count',
        'min', 'max', 'avg', 'sum', 'exists', 'get_bindings'
    ]

    def __init__(self, query):
        """
        Constructor

        :param query: The underlying query builder
        :type query: QueryBuilder
        """
        self._query = query

        self._model = None
        self._eager_load = {}
        self._macros = []

        self._on_delete = None

    def find(self, id, columns=None):
        """
        Find a model by its primary key

        :param id: The primary key value
        :type id: mixed

        :param columns: The columns to retrieve
        :type columns: list

        :return: The found model
        :rtype: eloquent.Model
        """
        if columns is None:
            columns = ['*']

        if isinstance(id, list):
            return self.find_many(id, columns)

        self._query.where(self._model.get_qualified_key_name(), '=', id)

        return self.first(columns)

    def find_many(self, id, columns=None):
        """
        Find a model by its primary key

        :param id: The primary key values
        :type id: list

        :param columns: The columns to retrieve
        :type columns: list

        :return: The found model
        :rtype: eloquent.Collection
        """
        if columns is None:
            columns = ['*']

        if not id:
            return self._model.new_collection()

        self._query.where_in(self._model.get_qualified_key_name(), id)

        return self.get(columns)

    def find_or_fail(self, id, columns=None):
        """
        Find a model by its primary key or raise an exception

        :param id: The primary key value
        :type id: mixed

        :param columns: The columns to retrieve
        :type columns: list

        :return: The found model
        :rtype: eloquent.Model

        :raises: ModelNotFound
        """
        result = self.find(id, columns)

        if isinstance(id, list):
            if len(result) == len(set(id)):
                return result
        elif result:
            return result

        raise ModelNotFound(self._model.__class__)

    def first(self, columns=None):
        """
        Execute the query and get the first result

        :param columns: The columns to get
        :type columns: list

        :return: The result
        :rtype: mixed
        """
        if columns is None:
            columns = ['*']

        return self.take(1).get(columns).first()

    def first_or_fail(self, columns=None):
        """
        Execute the query and get the first result or raise an exception

        :param columns: The columns to get
        :type columns: list

        :return: The result
        :rtype: mixed
        """
        model = self.first(columns)

        if model is not None:
            return model

        raise ModelNotFound(self._model.__class__)

    def get(self, columns=None):
        """
        Execute the query as a "select" statement.

        :param columns: The columns to get
        :type columns: list

        :rtype: eloquent.Collection
        """
        models = self.get_models(columns)

        # If we actually found models we will also eager load any relationships that
        # have been specified as needing to be eager loaded, which will solve the
        # n+1 query issue for the developers to avoid running a lot of queries.
        if len(models) > 0:
            models = self.eager_load_relations(models)

        return self._model.new_collection(models)

    def pluck(self, column):
        """
        Pluck a single column from the database.

        :param column: THe column to pluck
        :type column: str

        :return: The column value
        :rtype: mixed
        """
        result = self.first([column])

        if result:
            return result[column]

    def chunk(self, count):
        """
        Chunk the results of the query

        :param count: The chunk size
        :type count: int

        :return: The current chunk
        :rtype: list
        """
        page = 1
        results = self.for_page(page, count).get()

        while results:
            yield results

            page += 1

            results = self.for_page(page, count).get()

    def lists(self, column, key=None):
        """
        Get a list with the values of a given column

        :param column: The column to get the values for
        :type column: str

        :param key: The key
        :type key: str

        :return: The list of values
        :rtype: list or dict
        """
        results = self._query.lists(column, key)

        if self._model.has_get_mutator(column):
            if isinstance(results, dict):
                for key, value in results.items():
                    fill = {column: value}

                    results[key] = self._model.new_from_builder(fill).column
            else:
                for i, value in enumerate(results):
                    fill = {column: value}

                    results[i] = self._model.new_from_builder(fill).column

        return results

    def paginate(self, per_page=None, columns=None):
        """
        Paginate the given query.

        :param per_page: The number of records per page
        :type per_page: int

        :param columns: The columns to return
        :type columns: list

        :return: The paginator
        """
        # TODO

    def simple_paginate(self, per_page=None, columns=None):
        """
        Paginate the given query.

        :param per_page: The number of records per page
        :type per_page: int

        :param columns: The columns to return
        :type columns: list

        :return: The paginator
        """
        # TODO

    def update(self, _values=None, **values):
        """
        Update a record in the database

        :param values: The values of the update
        :type values: dict

        :return: The number of records affected
        :rtype: int
        """
        if _values is not None:
            values.update(_values)

        return self._query.update(self._add_updated_at_column(values))

    def increment(self, column, amount=1, extras=None):
        """
        Increment a column's value by a given amount

        :param column: The column to increment
        :type column: str

        :param amount: The amount by which to increment
        :type amount: int

        :param extras: Extra columns
        :type extras: dict

        :return: The number of rows affected
        :rtype: int
        """
        extras = self._add_updated_at_column(extras)

        return self._query.increment(column, amount, extras)

    def decrement(self, column, amount=1, extras=None):
        """
        Decrement a column's value by a given amount

        :param column: The column to increment
        :type column: str

        :param amount: The amount by which to increment
        :type amount: int

        :param extras: Extra columns
        :type extras: dict

        :return: The number of rows affected
        :rtype: int
        """
        extras = self._add_updated_at_column(extras)

        return self._query.decrement(column, amount, extras)

    def _add_updated_at_column(self, values):
        """
        Add the "updated_at" column to a dictionary of values.

        :param values: The values to update
        :type values: dict

        :return: The new dictionary of values
        :rtype: dict
        """
        if not self._model.uses_timestamps():
            return values

        column = self._model.get_updated_at_column()

        values.update({column: self._model.fresh_timestamp()})

        return values

    def delete(self):
        """
        Delete a record from the database.
        """
        if self._on_delete is not None:
            return self._on_delete(self)

        return self._query.delete()

    def force_delete(self):
        """
        Run the default delete function on the builder.
        """
        return self._query.delete()

    def on_delete(self, callback):
        """
        Register a replacement for the default delete function.

        :param callback: A replacement for the default delete function
        :type callback: callable
        """
        self._on_delete = callback

    def get_models(self, columns=None):
        """
        Get the hydrated models without eager loading.

        :param columns: The columns to get
        :type columns: list

        :return: A list of models
        :rtype: list
        """
        results = self._query.get(columns)

        connection = self._model.get_connection_name()

        return self._model.hydrate(results, connection).all()

    def eager_load_relations(self, models):
        """
        Eager load the relationship of the models.

        :param models:
        :type models: list

        :return: The models
        :rtype: list
        """
        for name, constraints in self._eager_load.items():
            if name.find('.') == -1:
                models = self._load_relation(models, name, constraints)

        return models

    def _load_relation(self, models, name, constraints):
        """
        Eagerly load the relationship on a set of models.

        :rtype: list
        """
        relation = self.get_relation(name)

        relation.add_eager_constraints(models)

        if callable(constraints):
            constraints(relation)
        else:
            relation.merge_query(constraints)

        models = relation.init_relation(models, name)

        results = relation.get_eager()

        return relation.match(models, results, name)

    def get_relation(self, relation):
        """
        Get the relation instance for the given relation name.

        :rtype: eloquent.orm.relations.Relation
        """
        from .relations import Relation

        query = Relation.no_constraints(lambda: getattr(self.get_model(), relation)())

        nested = self._nested_relations(relation)

        if len(nested) > 0:
            query.get_query().with_(nested)

        return query

    def _nested_relations(self, relation):
        """
        Get the deeply nested relations for a given top-level relation.

        :rtype: dict
        """
        nested = {}

        for name, constraints in self._eager_load.items():
            if self._is_nested(name, relation):
                nested[name[len(relation + '.'):]] = constraints

        return nested

    def _is_nested(self, name, relation):
        """
        Determine if the relationship is nested.

        :type name: str
        :type relation: str

        :rtype: bool
        """
        dots = name.find('.')

        return dots and name.startswith(relation + '.')

    def where(self, column, operator=Null(), value=None, boolean='and'):
        """
        Add a where clause to the query

        :param column: The column of the where clause, can also be a QueryBuilder instance for sub where
        :type column: str|Builder

        :param operator: The operator of the where clause
        :type operator: str

        :param value: The value of the where clause
        :type value: mixed

        :param boolean: The boolean of the where clause
        :type boolean: str

        :return: The current Builder instance
        :rtype: Builder
        """
        if isinstance(column, Builder):
            self._query.add_nested_where_query(column.get_query(), boolean)
        else:
            self._query.where(column, operator, value, boolean)

        return self

    def or_where(self, column, operator=None, value=None):
        """
        Add an "or where" clause to the query.

        :param column: The column of the where clause, can also be a QueryBuilder instance for sub where
        :type column: str or Builder

        :param operator: The operator of the where clause
        :type operator: str

        :param value: The value of the where clause
        :type value: mixed

        :return: The current Builder instance
        :rtype: Builder
        """
        return self.where(column, operator, value, 'or')

    def has(self, relation, operator='>=', count=1, boolean='and', extra=None):
        """
        Add a relationship count condition to the query.

        :param relation: The relation to count
        :type relation: str

        :param operator: The operator
        :type operator: str

        :param count: The count
        :type count: int

        :param boolean: The boolean value
        :type boolean: str

        :param extra: The extra query
        :type extra: Builder or callable

        :type: Builder
        """
        if relation.find('.') >= 0:
            return self._has_nested(relation, operator, count, boolean, extra)

        relation = self._get_has_relation_query(relation)

        query = relation.get_relation_count_query(relation.get_related().new_query(), self)

        # TODO: extra query
        if extra:
            if callable(extra):
                extra(query)

        return self._add_has_where(query, relation, operator, count, boolean)

    def _has_nested(self, relations, operator='>=', count=1, boolean='and', extra=None):
        """
        Add nested relationship count conditions to the query.

        :param relations: nested relations
        :type relations: str

        :param operator: The operator
        :type operator: str

        :param count: The count
        :type count: int

        :param boolean: The boolean value
        :type boolean: str

        :param extra: The extra query
        :type extra: Builder or callable

        :rtype: Builder
        """
        relations = relations.split('.')

        def closure(q):
            if len(relations) > 1:
                q.where_has(relations.pop(0), closure)
            else:
                q.has(relations.pop(0), operator, count, boolean, extra)

        return self.where_has(relations.pop(0), closure)

    def doesnt_have(self, relation, boolean='and', extra=None):
        """
        Add a relationship count to the query.

        :param relation: The relation to count
        :type relation: str

        :param boolean: The boolean value
        :type boolean: str

        :param extra: The extra query
        :type extra: Builder or callable

        :rtype: Builder
        """
        return self.has(relation, '<', 1, boolean, extra)

    def where_has(self, relation, extra, operator='>=', count=1):
        """
        Add a relationship count condition to the query with where clauses.

        :param relation: The relation to count
        :type relation: str

        :param extra: The extra query
        :type extra: Builder or callable

        :param operator: The operator
        :type operator: str

        :param count: The count
        :type count: int

        :rtype: Builder
        """
        return self.has(relation, operator, count, 'and', extra)

    def where_doesnt_have(self, relation, extra=None):
        """
        Add a relationship count condition to the query with where clauses.

        :param relation: The relation to count
        :type relation: str

        :param extra: The extra query
        :type extra: Builder or callable

        :rtype: Builder
        """
        return self.doesnt_have(relation, 'and', extra)

    def or_has(self, relation, operator='>=', count=1):
        """
        Add a relationship count condition to the query with an "or".

        :param relation: The relation to count
        :type relation: str

        :param operator: The operator
        :type operator: str

        :param count: The count
        :type count: int

        :rtype: Builder
        """
        return self.has(relation, operator, count, 'or')

    def or_where_has(self, relation, extra, operator='>=', count=1):
        """
        Add a relationship count condition to the query with where clauses and an "or".

        :param relation: The relation to count
        :type relation: str

        :param extra: The extra query
        :type extra: Builder or callable

        :param operator: The operator
        :type operator: str

        :param count: The count
        :type count: int

        :rtype: Builder
        """
        return self.has(relation, operator, count, 'or', extra)

    def _add_has_where(self, has_query, relation, operator, count, boolean):
        """
        Add the "has" condition where clause to the query.

        :param has_query: The has query
        :type has_query: Builder

        :param relation: The relation to count
        :type relation: eloquent.orm.relations.Relation

        :param operator: The operator
        :type operator: str

        :param count: The count
        :type count: int

        :param boolean: The boolean value
        :type boolean: str

        :rtype: Builder
        """
        self._merge_wheres_to_has(has_query, relation)

        if isinstance(count, basestring) and count.isdigit():
            count = QueryExpression(count)

        return self.where(QueryExpression('(%s)' % has_query.to_sql()), operator, count, boolean)

    def _merge_wheres_to_has(self, has_query, relation):
        """
        Merge the "wheres" from the relation query to a has query.

        :param has_query: The has query
        :type has_query: Builder

        :param relation: The relation to count
        :type relation: eloquent.orm.relations.Relation
        """
        relation_query = relation.get_base_query()

        has_query.merge_wheres(relation_query.wheres, relation_query.get_bindings())

        self._query.merge_bindings(has_query.get_query())

    def _get_has_relation_query(self, relation):
        """
        Get the "has" relation base query

        :type relation: str

        :rtype: Builder
        """
        from .relations import Relation

        return Relation.no_constraints(
            lambda: getattr(self.get_model(), relation)()
        )

    def with_(self, *relations):
        """
        Set the relationships that should be eager loaded.

        :return: The current Builder instance
        :rtype: Builder
        """
        if not relations:
            return self

        eagers = self._parse_relations(list(relations))

        self._eager_load.update(eagers)

        return self

    def _parse_relations(self, relations):
        """
        Parse a list of relations into individuals.

        :param relations: The relation to parse
        :type relations: list

        :rtype: dict
        """
        results = {}

        for relation in relations:
            if isinstance(relation, dict):
                name = list(relation.keys())[0]
                constraints = relation[name]
            else:
                name = relation
                constraints = self.__class__(self.get_query().new_query())

            results = self._parse_nested(name, results)

            results[name] = constraints

        return results

    def _parse_nested(self, name, results):
        """
        Parse the nested relationship in a relation.

        :param name: The name of the relationship
        :type name: str

        :type results: dict

        :rtype: dict
        """
        progress = []

        for segment in name.split('.'):
            progress.append(segment)

            last = '.'.join(progress)
            if last not in results:
                results[last] = self.__class__(self.get_query().new_query())

        return results

    def get_query(self):
        """
        Get the underlying query instance.

        :rtype: QueryBuilder
        """
        return self._query

    def set_query(self, query):
        """
        Set the underlying query instance.

        :param query: A QueryBuilder instance
        :type query: QueryBuilder
        """
        self._query = query

    def get_eager_loads(self):
        """
        Get the relationships being eager loaded.

        :rtype: dict
        """
        return self._eager_load

    def set_eager_loads(self, eager_load):
        """
        Sets the relationships to eager load.

        :type eager_load: dict

        :rtype: Builder
        """
        self._eager_load = eager_load

        return self

    def get_model(self):
        """
        Get the model instance of the model being queried

        :rtype: eloquent.Model
        """
        return self._model

    def set_model(self, model):
        """
        Set a model instance for the model being queried.

        :param model: The model instance
        :type model: eloquent.orm.Model

        :return: The current Builder instance
        :rtype: Builder
        """
        self._model = model

        self._query.from_(model.get_table())

        return self

    def __dynamic(self, method):
        attribute = getattr(self._query, method)

        def call(*args, **kwargs):
            result = attribute(*args, **kwargs)

            if method in self._passthru:
                return result
            else:
                return self

        if not callable(attribute):
            return attribute

        return call

    def __getattr__(self, item, *args):
        try:
            object.__getattribute__(self, item)
        except AttributeError:
            # TODO: macros and scopes
            return self.__dynamic(item)



# -*- coding: utf-8 -*-

from .expression import QueryExpression


class JoinClause(object):

    def __init__(self, table, type='inner'):
        self.type = type
        self.table = table

        self.clauses = []
        self.bindings = []

    def on(self, first, operator, second, boolean='and', where=False):
        self.clauses.append({
            'first': first,
            'operator': operator,
            'second': second,
            'boolean': boolean,
            'where': where
        })

        if where:
            self.bindings.append(second)

        return self

    def or_on(self, first, operator, second):
        return self.on(first, operator, second, 'or')

    def where(self, first, operator, second, boolean='and'):
        return self.on(first, operator, second, boolean, True)

    def or_where(self, first, operator, second):
        return self.where(first, operator, second, 'or')

    def where_null(self, column, boolean='and'):
        return self.on(column, 'IS', QueryExpression('NULL'), boolean, False)

    def or_where_null(self, column):
        return self.where_null(column, 'or')

    def where_not_null(self, column, boolean='and'):
        return self.on(column, 'IS', QueryExpression('NOT NULL'), boolean, False)

    def or_where_not_null(self, column):
        return self.where_not_null(column, 'or')



# -*- coding: utf-8 -*-

import arrow
from flexmock import flexmock, flexmock_teardown
from ... import EloquentTestCase

from eloquent.query.builder import QueryBuilder
from eloquent.query.grammars import QueryGrammar
from eloquent.query.expression import QueryExpression
from eloquent.orm.builder import Builder
from eloquent.orm.model import Model
from eloquent.orm.relations import MorphOne, MorphMany
from eloquent.orm.collection import Collection


class OrmMorphTestCase(EloquentTestCase):

    def tearDown(self):
        flexmock_teardown()

    def test_morph_one_sets_proper_constraints(self):
        self._get_one_relation()

    def test_morph_one_eager_constraints_are_properly_added(self):
        relation = self._get_one_relation()
        relation.get_query().get_query().should_receive('where_in').once().with_args('table.morph_id', [1, 2])
        relation.get_query().should_receive('where').once()\
            .with_args('table.morph_type', relation.get_parent().__class__.__name__)

        model1 = Model()
        model1.id = 1
        model2 = Model()
        model2.id = 2
        relation.add_eager_constraints([model1, model2])

    def test_morph_many_sets_proper_constraints(self):
        self._get_many_relation()

    def test_morph_many_eager_constraints_are_properly_added(self):
        relation = self._get_many_relation()
        relation.get_query().get_query().should_receive('where_in').once().with_args('table.morph_id', [1, 2])
        relation.get_query().should_receive('where').once()\
            .with_args('table.morph_type', relation.get_parent().__class__.__name__)

        model1 = Model()
        model1.id = 1
        model2 = Model()
        model2.id = 2
        relation.add_eager_constraints([model1, model2])

    def test_create(self):
        relation = self._get_one_relation()
        created = flexmock(Model())
        created.should_receive('set_attribute').once().with_args('morph_id', 1)
        created.should_receive('set_attribute').once()\
            .with_args('morph_type', relation.get_parent().__class__.__name__)
        relation.get_related().should_receive('new_instance').once().with_args({'name': 'john'}).and_return(created)
        created.should_receive('save').once().and_return(True)

        self.assertEqual(created, relation.create(name='john'))

    def test_find_or_new_finds_model(self):
        relation = self._get_one_relation()
        model = flexmock()
        model.foo = 'bar'
        relation.get_query().should_receive('find').once().with_args('foo', ['*']).and_return(model)
        relation.get_related().should_receive('new_instance').never()
        model.should_receive('set_attribute').never()
        model.should_receive('save').never()

        self.assertEqual('bar', relation.find_or_new('foo').foo)

    def test_find_or_new_returns_new_model_with_morph_keys_set(self):
        relation = self._get_one_relation()
        model = flexmock()
        model.foo = 'bar'
        relation.get_query().should_receive('find').once().with_args('foo', ['*']).and_return(None)
        relation.get_related().should_receive('new_instance').once().with_args().and_return(model)
        model.should_receive('set_attribute').once().with_args('morph_id', 1)
        model.should_receive('set_attribute').once().with_args('morph_type', relation.get_parent().__class__.__name__)
        model.should_receive('save').never()

        self.assertEqual('bar', relation.find_or_new('foo').foo)

    def test_first_or_new_returns_first_model(self):
        relation = self._get_one_relation()
        model = flexmock()
        model.foo = 'bar'
        relation.get_query().should_receive('where').once().with_args({'foo': 'bar'}).and_return(relation.get_query())
        relation.get_query().should_receive('first').once().with_args().and_return(model)
        relation.get_related().should_receive('new_instance').never()
        model.should_receive('set_attribute').never()
        model.should_receive('set_attribute').never()
        model.should_receive('save').never()

        self.assertEqual('bar', relation.first_or_new({'foo': 'bar'}).foo)

    def test_first_or_new_returns_new_model_with_morph_keys_set(self):
        relation = self._get_one_relation()
        model = flexmock()
        model.foo = 'bar'
        relation.get_query().should_receive('where').once().with_args({'foo': 'bar'}).and_return(relation.get_query())
        relation.get_query().should_receive('first').once().with_args().and_return(None)
        relation.get_related().should_receive('new_instance').once().with_args().and_return(model)
        model.should_receive('set_attribute').once().with_args('morph_id', 1)
        model.should_receive('set_attribute').once().with_args('morph_type', relation.get_parent().__class__.__name__)
        model.should_receive('save').never()

        self.assertEqual('bar', relation.first_or_new({'foo': 'bar'}).foo)

    def test_first_or_create_returns_first_model(self):
        relation = self._get_one_relation()
        model = flexmock()
        model.foo = 'bar'
        relation.get_query().should_receive('where').once().with_args({'foo': 'bar'}).and_return(relation.get_query())
        relation.get_query().should_receive('first').once().with_args().and_return(model)
        relation.get_related().should_receive('new_instance').never()
        model.should_receive('set_attribute').never()
        model.should_receive('set_attribute').never()
        model.should_receive('save').never()

        self.assertEqual('bar', relation.first_or_create({'foo': 'bar'}).foo)

    def test_first_or_create_creates_new_morph_model(self):
        relation = self._get_one_relation()
        model = flexmock()
        model.foo = 'bar'
        relation.get_query().should_receive('where').once().with_args({'foo': 'bar'}).and_return(relation.get_query())
        relation.get_query().should_receive('first').once().with_args().and_return(None)
        relation.get_related().should_receive('new_instance').once().with_args({'foo': 'bar'}).and_return(model)
        model.should_receive('set_attribute').once().with_args('morph_id', 1)
        model.should_receive('set_attribute').once().with_args('morph_type', relation.get_parent().__class__.__name__)
        model.should_receive('save').once().and_return()

        self.assertEqual('bar', relation.first_or_create({'foo': 'bar'}).foo)

    def test_update_or_create_finds_first_model_and_updates(self):
        relation = self._get_one_relation()
        model = flexmock()
        model.foo = 'bar'
        relation.get_query().should_receive('where').once().with_args({'foo': 'bar'}).and_return(relation.get_query())
        relation.get_query().should_receive('first').once().with_args().and_return(model)
        relation.get_related().should_receive('new_instance').never()
        model.should_receive('set_attribute').never()
        model.should_receive('set_attribute').never()
        model.should_receive('fill').once().with_args({'bar': 'baz'})
        model.should_receive('save').once().and_return(True)

        self.assertEqual('bar', relation.update_or_create({'foo': 'bar'}, {'bar': 'baz'}).foo)

    def test_update_or_create_finds_creates_new_morph_model(self):
        relation = self._get_one_relation()
        model = flexmock()
        model.foo = 'bar'
        relation.get_query().should_receive('where').once().with_args({'foo': 'bar'}).and_return(relation.get_query())
        relation.get_query().should_receive('first').once().with_args().and_return(None)
        relation.get_related().should_receive('new_instance').once().with_args().and_return(model)
        model.should_receive('set_attribute').once().with_args('morph_id', 1)
        model.should_receive('set_attribute').once().with_args('morph_type', relation.get_parent().__class__.__name__)
        model.should_receive('fill').once().with_args({'bar': 'baz'})
        model.should_receive('save').once().and_return(True)

        self.assertEqual('bar', relation.update_or_create({'foo': 'bar'}, {'bar': 'baz'}).foo)

    def _get_many_relation(self):
        flexmock(Builder)
        query = flexmock(QueryBuilder(None, QueryGrammar(), None))
        builder = Builder(query)
        builder.should_receive('where').with_args('table.morph_id', '=', 1)
        related = flexmock(Model())
        builder.should_receive('get_model').and_return(related)
        parent = flexmock(Model())
        parent.should_receive('get_attribute').with_args('id').and_return(1)
        parent.should_receive('get_morph_class').and_return(parent.__class__.__name__)
        builder.should_receive('where').once().with_args('table.morph_type', parent.__class__.__name__)

        return MorphMany(builder, parent, 'table.morph_type', 'table.morph_id', 'id')

    def _get_one_relation(self):
        flexmock(Builder)
        query = flexmock(QueryBuilder(None, QueryGrammar(), None))
        builder = Builder(query)
        builder.should_receive('where').with_args('table.morph_id', '=', 1)
        related = flexmock(Model())
        builder.should_receive('get_model').and_return(related)
        parent = flexmock(Model())
        parent.should_receive('get_attribute').with_args('id').and_return(1)
        parent.should_receive('get_morph_class').and_return(parent.__class__.__name__)
        builder.should_receive('where').once().with_args('table.morph_type', parent.__class__.__name__)

        return MorphOne(builder, parent, 'table.morph_type', 'table.morph_id', 'id')



# -*- coding: utf-8 -*-


class ModelNotFound(RuntimeError):

    def __init__(self, model):
        self._model = model

        self.message = 'No query results found for model [%s]' % self._model.__name__

    def __str__(self):
        return self.message


class MassAssignmentError(RuntimeError):
    pass



# -*- coding: utf-8 -*-

import time
import logging
from contextlib import contextmanager
from .connection_interface import ConnectionInterface
from ..query.grammars.grammar import QueryGrammar
from ..query.builder import QueryBuilder
from ..query.expression import QueryExpression
from ..query.processors.processor import QueryProcessor
from ..exceptions.query import QueryException


query_logger = logging.getLogger('eloquent.connection.queries')


class Connection(ConnectionInterface):

    def __init__(self, connection, database='', table_prefix='', config=None):
        """
        :param connection: A dbapi connection instance
        :type connection: mixed

        :param database: The database name
        :type database: str

        :param table_prefix: The table prefix
        :type table_prefix: str

        :param config: The connection configuration
        :type config: dict
        """
        self._connection = connection
        self._cursor = None

        self._read_connection = None

        self._database = database

        self._table_prefix = table_prefix

        if config is None:
            config = {}

        self._config = config

        self._reconnector = None

        self._transactions = 0

        self._pretending = False

        self._logging_queries = config.get('log_queries', False)

        self._query_grammar = self.get_default_query_grammar()

        self._schema_grammar = None

        self._post_processor = self.get_default_post_processor()

        self.use_default_query_grammar()

    def use_default_query_grammar(self):
        self._query_grammar = self.get_default_query_grammar()

    def get_default_query_grammar(self):
        return QueryGrammar()

    def use_default_post_processor(self):
        self._post_processor = self.get_default_post_processor()

    def get_default_post_processor(self):
        return QueryProcessor()

    def table(self, table):
        """
        Begin a fluent query against a database table

        :param table: The database table
        :type table: str

        :return: A QueryBuilder instance
        :rtype: QueryBuilder
        """
        query = self.query()

        return query.from_(table)

    def query(self):
        """
        Begin a fluent query

        :return: A QueryBuilder instance
        :rtype: QueryBuilder
        """
        query = QueryBuilder(self, self._query_grammar, self._post_processor)

        return query

    def raw(self, value):
        return QueryExpression(value)

    def select_one(self, query, bindings=None):
        if bindings is None:
            bindings = {}

        records = self.select(query, bindings)

        if len(records):
            return records[1]

        return None

    def select_from_write_connection(self, query, bindings=None):
        if bindings is None:
            bindings = {}

        return self.select(query, bindings)

    def select(self, query, bindings=None, use_read_connection=True):
        def callback(me, query_, bindings_):
            if me.pretending():
                return []

            bindings_ = me.prepare_bindings(bindings_)
            cursor = self._get_cursor_for_select(use_read_connection)
            cursor.execute(query_, bindings_)

            return cursor.fetchall()

        return self._run(query, bindings, callback)

    def _get_cursor_for_select(self, use_read_connection=True):
        if use_read_connection:
            self._cursor = self.get_read_connection().cursor()
        else:
            self._cursor = self.get_connection().cursor()

        return self._cursor

    def insert(self, query, bindings=None):
        return self.statement(query, bindings)

    def update(self, query, bindings=None):
        return self.affecting_statement(query, bindings)

    def delete(self, query, bindings=None):
        return self.affecting_statement(query, bindings)

    def statement(self, query, bindings=None):
        def callback(me, query_, bindings_):
            if me.pretending():
                return True

            bindings_ = me.prepare_bindings(bindings_)

            return me._new_cursor().execute(query_, bindings_)

        return self._run(query, bindings, callback)

    def affecting_statement(self, query, bindings=None):
        def callback(me, query_, bindings_):
            if me.pretending():
                return True

            bindings_ = me.prepare_bindings(bindings_)

            cursor = me._new_cursor()
            cursor.execute(query_, bindings_)

            return cursor.rowcount

        return self._run(query, bindings, callback)

    def _new_cursor(self):
        self._cursor = self.get_connection().cursor()

        return self._cursor

    def get_cursor(self):
        return self._cursor

    def unprepared(self, query):
        def callback(me, query_, _):
            if me.pretending():
                return True

            return bool(me.get_connection().execute(query_))

        return self._run(query, {}, callback)

    def prepare_bindings(self, bindings):
        if bindings is None:
            return []

        return bindings

    @contextmanager
    def transaction(self):
        self.begin_transaction()

        try:
            yield self
        except Exception as e:
            self.rollback()
            raise

        try:
            self.commit()
        except Exception:
            self.rollback()
            raise

    def begin_transaction(self):
        self._transactions += 1

    def commit(self):
        if self._transactions == 1:
            self._connection.commit()

        self._transactions -= 1

    def rollback(self):
        if self._transactions == 1:
            self._transactions = 0

            self._connection.rollback()
        else:
            self._transactions -= 1

    def transaction_level(self):
        return self._transactions

    def pretend(self, callback):
        logging_queries = self._logging_queries

        self.enable_query_log()

        self._pretending = True

        callback(self)

        self._pretending = False

        self._logging_queries = logging_queries

    def _run(self, query, bindings, callback):
        self._reconnect_if_missing_connection()

        start = time.time()

        try:
            result = self._run_query_callback(query, bindings, callback)
        except QueryException as e:
            result = self._try_again_if_caused_by_lost_connection(
                e, query, bindings, callback
            )

        t = self._get_elapsed_time(start)

        self.log_query(query, bindings, t)

        return result

    def _run_query_callback(self, query, bindings, callback):
        try:
            result = callback(self, query, bindings)
        except Exception as e:
            raise QueryException(query, self.prepare_bindings(bindings), e)

        return result

    def _try_again_if_caused_by_lost_connection(self, e, query,
                                                bindings, callback):
        if self._caused_by_lost_connection(e):
            self.reconnect()

            return self._run_query_callback(query, bindings, callback)

        raise e

    def _caused_by_lost_connection(self, e):
        message = e.message

        for s in ['server has gone away',
                  'no connection to the server',
                  'Lost Connection']:
            if s in message:
                return True

        return False

    def disconnect(self):
        if self._connection:
            self._connection.close()

        if self._read_connection:
            self._read_connection.close()

        self.set_connection(None).set_read_connection(None)

    def reconnect(self):
        if self._reconnector is not None and callable(self._reconnector):
            return self._reconnector(self)

        raise Exception('Lost connection and no reconnector available')

    def _reconnect_if_missing_connection(self):
        if self.get_connection() is None or self.get_read_connection() is None:
            self.reconnect()

    def log_query(self, query, bindings, time_=None):
        if not self._logging_queries:
            return

        query = self._get_cursor_query(query, bindings)

        if query:
            log = 'Executed %s' % (query,)

            if time_:
                log += ' in %sms' % time_

            query_logger.debug(log,
                               extra={
                                   'query': query,
                                   'bindings': bindings,
                                   'elapsed_time': time_
                               })

    def _get_elapsed_time(self, start):
        return round((time.time() - start) * 1000, 2)

    def _get_cursor_query(self, query, bindings):
        return query, bindings

    def get_connection(self):
        return self._connection

    def get_read_connection(self):
        if self._transactions >= 1:
            return self.get_connection()

        if self._read_connection is not None:
            return self._read_connection

        return self._connection

    def set_connection(self, connection):
        if self._transactions >= 1:
            raise RuntimeError("Can't swap dbapi connection"
                               "while within transaction.")

        self._connection = connection

        return self

    def set_read_connection(self, connection):
        self._read_connection = connection

        return self

    def set_reconnector(self, reconnector):
        self._reconnector = reconnector

        return self

    def get_name(self):
        return self._config.get('name')

    def get_query_grammar(self):
        return self._query_grammar

    def set_query_grammar(self, grammar):
        self._query_grammar = grammar

    def get_schema_grammar(self):
        return self._schema_grammar

    def set_schema_grammar(self, grammar):
        self._schema_grammar = grammar

    def get_post_processor(self):
        """
        Get the query post processor used by the connection

        :return: The query post processor
        :rtype: QueryProcessor
        """
        return self._post_processor

    def set_post_processor(self, processor):
        """
        Set the query post processor used by the connection

        :param processor: The query post processor
        :type processor: QueryProcessor
        """
        self._post_processor = processor

    def pretending(self):
        return self._pretending

    def enable_query_log(self):
        self._logging_queries = True

    def disable_query_log(self):
        self._logging_queries = False

    def logging(self):
        return self._logging_queries

    def get_database_name(self):
        return self._database

    def get_table_prefix(self):
        return self._table_prefix

    def set_table_prefix(self, prefix):
        self._table_prefix = prefix

        self.get_query_grammar().set_table_prefix(prefix)

    def with_table_prefix(self, grammar):
        grammar.set_table_prefix(self._table_prefix)

        return grammar

    def __enter__(self):
        self.begin_transaction()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            try:
                self.commit()
            except Exception:
                self.rollback()
                raise
        else:
            self.rollback()
            raise (exc_type, exc_val, exc_tb)



# -*- coding: utf-8 -*-


class ConnectionInterface(object):

    def table(self, table):
        """
        Begin a fluent query against a database table

        :param table: The database table
        :type table: str

        :return: A QueryBuilder instance
        :rtype: QueryBuilder
        """
        raise NotImplementedError()

    def query(self):
        """
        Begin a fluent query

        :return: A QueryBuilder instance
        :rtype: QueryBuilder
        """
        raise NotImplementedError()

    def raw(self, value):
        """
        Get a new raw query expression

        :param value: The value
        :type value: mixed

        :return: A QueryExpression instance
        :rtype: QueryExpression
        """
        raise NotImplementedError()

    def select_one(self, query, bindings=None):
        """
        Run a select statement and return a single result

        :param query: The select statement
        :type query: str
        :param bindings: The query bindings
        :type bindings: dict

        :return: mixed
        """
        raise NotImplementedError()

    def select(self, query, bindings=None):
        """
        Run a select statement against the database

        :param query: The select statement
        :type query: str
        :param bindings: The query bindings
        :type bindings: dict

        :return: mixed
        """
        raise NotImplementedError()

    def insert(self, query, bindings=None):
        """
        Run an insert statement against the database

        :param query: The insert statement
        :type query: str
        :param bindings: The query bindings
        :type bindings: dict

        :return: mixed
        """
        raise NotImplementedError()

    def update(self, query, bindings=None):
        """
        Run an update statement against the database

        :param query: The update statement
        :type query: str
        :param bindings: The query bindings
        :type bindings: dict

        :return: mixed
        """
        raise NotImplementedError()

    def delete(self, query, bindings=None):
        """
        Run a delete statement against the database

        :param query: The select statement
        :type query: str
        :param bindings: The query bindings
        :type bindings: dict

        :return: mixed
        """
        raise NotImplementedError()

    def statement(self, query, bindings=None):
        """
        Run an SQL statement and return the boolean result

        :param query: The select statement
        :type query: str
        :param bindings: The query bindings
        :type bindings: dict

        :return: Boolean result
        :rtype: bool
        """
        raise NotImplementedError()

    def affecting_statement(self, query, bindings=None):
        """
        Run an SQL statement and return the number of affected rows

        :param query: The select statement
        :type query: str
        :param bindings: The query bindings
        :type bindings: dict

        :return: Number of affected rows
        :rtype: int
        """
        raise NotImplementedError()

    def unprepared(self, query):
        """
        Run a raw, unprepared query against the dbapi connection

        :param query: The raw query
        :type query: str

        :return: Boolean result
        :rtype: bool
        """
        raise NotImplementedError()

    def prepare_bindings(self, bindings):
        """
        Prepare the query bindings for execution

        :param bindings: The query bindings
        :type bindings: dict

        :return: The prepared bindings
        :rtype: dict
        """
        raise NotImplementedError()

    def transaction(self):
        raise NotImplementedError()

    def begin_transaction(self):
        raise NotImplementedError()

    def commit(self):
        raise NotImplementedError()

    def rollback(self):
        raise NotImplementedError()

    def transaction_level(self):
        raise NotImplementedError()

    def pretend(self, callback):
        raise NotImplementedError()





# -*- coding: utf-8 -*-

import sqlite3

from .connector import Connector


class DictCursor(object):

    def __init__(self, cursor, row):
        self.dict = {}
        self.cursor = cursor

        for idx, col in enumerate(cursor.description):
            self.dict[col[0]] = row[idx]

    def __getattr__(self, item):
        try:
            return object.__getattribute__(self, item)
        except AttributeError:
            return getattr(self.cursor, item)

    def __getitem__(self, item):
        return self.dict[item]

    def items(self):
        return self.dict.items()


class SQLiteConnector(Connector):

    def connect(self, config):
        connection = self.get_api().connect(**self.get_config(config))
        connection.isolation_level = None
        connection.row_factory = DictCursor

        return connection

    def get_api(self):
        return sqlite3



# -*- coding: utf-8 -*-

try:
    import psycopg2
    import psycopg2.extras
except ImportError:
    pass

from .connector import Connector


class PostgresConnector(Connector):

    def connect(self, config):
        connection = self.get_api().connect(
            connection_factory=psycopg2.extras.DictConnection,
            **self.get_config(config)
        )
        connection.autocommit = True

        return connection

    def get_api(self):
        return psycopg2



# -*- coding: utf-8 -*-

import threading
from .connections.connection_resolver_interface import ConnectionResolverInterface
from .connectors.connection_factory import ConnectionFactory
from .exceptions import ArgumentError


class BaseDatabaseManager(ConnectionResolverInterface):

    def __init__(self, config, factory=ConnectionFactory()):
        """
        :param config: The connections configuration
        :type config: dict

        :param factory: A connection factory
        :type factory: ConnectionFactory
        """
        self._config = config
        self._factory = factory

        self._connections = {}

        self._extensions = {}

    def connection(self, name=None):
        """
        Get a database connection instance

        :param name: The connection name
        :type name: str

        :return: A Connection instance
        :rtype: eloquent.connections.Connection
        """
        name, type = self._parse_connection_name(name)

        if name not in self._connections:
            connection = self._make_connection(name)

            self._set_connection_for_type(connection, type)

            self._connections[name] = self._prepare(connection)

        return self._connections[name]

    def _parse_connection_name(self, name):
        """
        Parse the connection into a tuple of the name and read / write type

        :param name: The name of the connection
        :type name: str

        :return: A tuple of the name and read / write type
        :rtype: tuple
        """
        if name is None:
            name = self.get_default_connection()

        if name.endswith(('::read', '::write')):
            return name.split('::', 1)

        return name, None

    def purge(self, name=None):
        """
        Disconnect from the given database and remove from local cache

        :param name: The name of the connection
        :type name: str

        :rtype: None
        """
        self.disconnect(name)

        if name in self._connections:
            del self._connections[name]

    def disconnect(self, name=None):
        if name is None:
            name = self.get_default_connection()

        if name in self._connections:
            self._connections[name].disconnect()

    def reconnect(self, name=None):
        if name is None:
            name = self.get_default_connection()

        self.disconnect(name)

        if name not in self._connections:
            return self.connection(name)

        return self._refresh_api_connections(name)

    def _refresh_api_connections(self, name):
        fresh = self._make_connection(name)

        return self._connections[name]\
            .set_connection(fresh.get_connection())\
            .set_read_connection(fresh.get_read_connection())

    def _make_connection(self, name):
        config = self._get_config(name)

        if name in self._extensions:
            return self._extensions[name](config, name)

        driver = config['driver']

        if driver in self._extensions:
            return self._extensions[driver](config, name)

        return self._factory.make(config, name)

    def _prepare(self, connection):
        def reconnector(connection_):
            self.reconnect(connection_.get_name())

        connection.set_reconnector(reconnector)

        return connection

    def _set_connection_for_type(self, connection, type):
        if type == 'read':
            connection.set_connection(connection.get_read_api())
        elif type == 'write':
            connection.set_read_connection(connection.get_api())

        return connection

    def _get_config(self, name):
        if name is None:
            name = self.get_default_connection()

        connections = self._config

        config = connections.get(name)
        if not config:
            raise ArgumentError('Database [%s] not configured' % name)

        return config

    def get_default_connection(self):
        if len(self._config) == 1:
            return list(self._config.keys())[0]

        return self._config['default']

    def set_default_connection(self, name):
        self._config['default'] = name

    def extend(self, name, resolver):
        self._extensions[name] = resolver

    def get_connections(self):
        return self._connections

    def __getattr__(self, item):
        try:
            return object.__getattribute__(self, item)
        except AttributeError as e:
            return getattr(self.connection(), item)


class DatabaseManager(BaseDatabaseManager, threading.local):

    pass



# -*- coding: utf-8 -*-

try:
    import MySQLdb as mysql
    from MySQLdb.cursors import DictCursor as cursor_class
    keys_fix = {
        'password': 'passwd',
        'database': 'db'
    }
except ImportError:
    try:
        import pymysql as mysql
        from pymysql.cursors import DictCursor as cursor_class
        keys_fix = {}
    except ImportError:
        mysql = None

from .connector import Connector


class MySqlConnector(Connector):

    def connect(self, config):
        config = dict(config.items())
        for key, value in keys_fix.items():
            config[value] = config[key]
            del config[key]

        config['autocommit'] = True
        config['cursorclass'] = cursor_class

        return self.get_api().connect(**self.get_config(config))

    def get_api(self):
        return mysql



# -*- coding: utf-8 -*-


import arrow
from flexmock import flexmock, flexmock_teardown
from ... import EloquentTestCase
from ...utils import MockConnection

from eloquent.query.builder import QueryBuilder
from eloquent.query.grammars import QueryGrammar
from eloquent.query.processors import QueryProcessor
from eloquent.query.expression import QueryExpression
from eloquent.orm.builder import Builder
from eloquent.orm.model import Model
from eloquent.orm.relations import BelongsToMany
from eloquent.orm.relations.pivot import Pivot
from eloquent.orm.collection import Collection


class OrmBelongsToTestCase(EloquentTestCase):

    def tearDown(self):
        flexmock_teardown()

    def test_models_are_properly_hydrated(self):
        model1 = OrmBelongsToManyModelStub()
        model1.fill(name='john', pivot_user_id=1, pivot_role_id=2)
        model2 = OrmBelongsToManyModelStub()
        model2.fill(name='jane', pivot_user_id=3, pivot_role_id=4)
        models = [model1, model2]

        base_builder = flexmock(Builder(QueryBuilder(MockConnection().prepare_mock(),
                                                     QueryGrammar(), QueryProcessor())))

        relation = self._get_relation()
        relation.get_parent().should_receive('get_connection_name').and_return('foo.connection')
        relation.get_query().get_query().should_receive('add_select').once()\
            .with_args(*['roles.*', 'user_role.user_id AS pivot_user_id', 'user_role.role_id AS pivot_role_id'])\
            .and_return(relation.get_query())
        relation.get_query().should_receive('get_models').once().and_return(models)
        relation.get_query().should_receive('eager_load_relations').once().with_args(models).and_return(models)
        relation.get_related().should_receive('new_collection').replace_with(lambda l: Collection(l))
        relation.get_query().should_receive('get_query').once().and_return(base_builder)

        results = relation.get()

        self.assertIsInstance(results, Collection)

        # Make sure the foreign keys were set on the pivot models
        self.assertEqual('user_id', results[0].pivot.get_foreign_key())
        self.assertEqual('role_id', results[0].pivot.get_other_key())

        self.assertEqual('john', results[0].name)
        self.assertEqual(1, results[0].pivot.user_id)
        self.assertEqual(2, results[0].pivot.role_id)
        self.assertEqual('foo.connection', results[0].pivot.get_connection_name())

        self.assertEqual('jane', results[1].name)
        self.assertEqual(3, results[1].pivot.user_id)
        self.assertEqual(4, results[1].pivot.role_id)
        self.assertEqual('foo.connection', results[1].pivot.get_connection_name())

        self.assertEqual('user_role', results[0].pivot.get_table())
        self.assertTrue(results[0].pivot.exists)

    def test_timestamps_can_be_retrieved_properly(self):
        model1 = OrmBelongsToManyModelStub()
        model1.fill(name='john', pivot_user_id=1, pivot_role_id=2)
        model2 = OrmBelongsToManyModelStub()
        model2.fill(name='jane', pivot_user_id=3, pivot_role_id=4)
        models = [model1, model2]

        base_builder = flexmock(Builder(QueryBuilder(MockConnection().prepare_mock(),
                                                     QueryGrammar(), QueryProcessor())))

        relation = self._get_relation().with_timestamps()
        relation.get_parent().should_receive('get_connection_name').and_return('foo.connection')
        relation.get_query().get_query().should_receive('add_select').once()\
            .with_args(
                'roles.*',
                'user_role.user_id AS pivot_user_id',
                'user_role.role_id AS pivot_role_id',
                'user_role.created_at AS pivot_created_at',
                'user_role.updated_at AS pivot_updated_at'
            )\
            .and_return(relation.get_query())
        relation.get_query().should_receive('get_models').once().and_return(models)
        relation.get_query().should_receive('eager_load_relations').once().with_args(models).and_return(models)
        relation.get_related().should_receive('new_collection').replace_with(lambda l: Collection(l))
        relation.get_query().should_receive('get_query').once().and_return(base_builder)

        results = relation.get()

    def test_models_are_properly_matched_to_parents(self):
        relation = self._get_relation()

        result1 = OrmBelongsToManyModelPivotStub()
        result1.pivot.user_id = 1

        result2 = OrmBelongsToManyModelPivotStub()
        result2.pivot.user_id = 2

        result3 = OrmBelongsToManyModelPivotStub()
        result3.pivot.user_id = 2

        model1 = OrmBelongsToManyModelStub()
        model1.id = 1
        model2 = OrmBelongsToManyModelStub()
        model2.id = 2
        model3 = OrmBelongsToManyModelStub()
        model3.id = 3

        relation.get_related().should_receive('new_collection').replace_with(lambda l: Collection(l))
        models = relation.match([model1, model2, model3], Collection([result1, result2, result3]), 'foo')

        self.assertEqual(1, models[0].foo[0].pivot.user_id)
        self.assertEqual(1, len(models[0].foo))
        self.assertEqual(2, models[1].foo[0].pivot.user_id)
        self.assertEqual(2, models[1].foo[1].pivot.user_id)
        self.assertEqual(2, len(models[1].foo))
        self.assertFalse(hasattr(models[2], 'foo'))

    def test_relation_is_properly_initialized(self):
        relation = self._get_relation()
        relation.get_related().should_receive('new_collection').replace_with(lambda l=None: Collection(l or []))
        model = flexmock(Model())
        model.should_receive('set_relation').once().with_args('foo', Collection)
        models = relation.init_relation([model], 'foo')

        self.assertEqual([model], models)

    def test_eager_constraints_are_properly_added(self):
        relation = self._get_relation()
        relation.get_query().get_query().should_receive('where_in').once().with_args('user_role.user_id', [1, 2])
        model1 = OrmBelongsToManyModelStub()
        model1.id = 1
        model2 = OrmBelongsToManyModelStub()
        model2.id = 2

        relation.add_eager_constraints([model1, model2])

    def test_attach_inserts_pivot_table_record(self):
        flexmock(BelongsToMany, touch_if_touching=lambda: True)
        relation = self._get_relation()
        query = flexmock()
        query.should_receive('from_').once().with_args('user_role').and_return(query)
        query.should_receive('insert').once().with_args([{'user_id': 1, 'role_id': 2, 'foo': 'bar'}]).and_return(True)
        mock_query_builder = flexmock()
        relation.get_query().should_receive('get_query').and_return(mock_query_builder)
        mock_query_builder.should_receive('new_query').once().and_return(query)
        relation.should_receive('touch_if_touching').once()

        relation.attach(2, {'foo': 'bar'})

    def test_attach_multiple_inserts_pivot_table_record(self):
        flexmock(BelongsToMany, touch_if_touching=lambda: True)
        relation = self._get_relation()
        query = flexmock()
        query.should_receive('from_').once().with_args('user_role').and_return(query)
        query.should_receive('insert').once().with_args(
            [
                {'user_id': 1, 'role_id': 2, 'foo': 'bar'},
                {'user_id': 1, 'role_id': 3, 'bar': 'baz', 'foo': 'bar'}
            ]
        ).and_return(True)
        mock_query_builder = flexmock()
        relation.get_query().should_receive('get_query').and_return(mock_query_builder)
        mock_query_builder.should_receive('new_query').once().and_return(query)
        relation.should_receive('touch_if_touching').once()

        relation.attach([2, {3: {'bar': 'baz'}}], {'foo': 'bar'})

    def test_attach_inserts_pivot_table_records_with_timestamps_when_ncessary(self):
        flexmock(BelongsToMany, touch_if_touching=lambda: True)
        relation = self._get_relation().with_timestamps()
        query = flexmock()
        query.should_receive('from_').once().with_args('user_role').and_return(query)
        now = arrow.get().naive
        query.should_receive('insert').once().with_args(
            [
                {'user_id': 1, 'role_id': 2, 'foo': 'bar', 'created_at': now, 'updated_at': now}
            ]
        ).and_return(True)
        mock_query_builder = flexmock()
        relation.get_query().should_receive('get_query').and_return(mock_query_builder)
        mock_query_builder.should_receive('new_query').once().and_return(query)
        relation.get_parent().should_receive('fresh_timestamp').once().and_return(now)
        relation.should_receive('touch_if_touching').once()

        relation.attach(2, {'foo': 'bar'})

    def test_attach_inserts_pivot_table_records_with_a_created_at_timestamp(self):
        flexmock(BelongsToMany, touch_if_touching=lambda: True)
        relation = self._get_relation().with_pivot('created_at')
        query = flexmock()
        query.should_receive('from_').once().with_args('user_role').and_return(query)
        now = arrow.get().naive
        query.should_receive('insert').once().with_args(
            [
                {'user_id': 1, 'role_id': 2, 'foo': 'bar', 'created_at': now}
            ]
        ).and_return(True)
        mock_query_builder = flexmock()
        relation.get_query().should_receive('get_query').and_return(mock_query_builder)
        mock_query_builder.should_receive('new_query').once().and_return(query)
        relation.get_parent().should_receive('fresh_timestamp').once().and_return(now)
        relation.should_receive('touch_if_touching').once()

        relation.attach(2, {'foo': 'bar'})

    def test_attach_inserts_pivot_table_records_with_an_updated_at_timestamp(self):
        flexmock(BelongsToMany, touch_if_touching=lambda: True)
        relation = self._get_relation().with_pivot('updated_at')
        query = flexmock()
        query.should_receive('from_').once().with_args('user_role').and_return(query)
        now = arrow.get().naive
        query.should_receive('insert').once().with_args(
            [
                {'user_id': 1, 'role_id': 2, 'foo': 'bar', 'updated_at': now}
            ]
        ).and_return(True)
        mock_query_builder = flexmock()
        relation.get_query().should_receive('get_query').and_return(mock_query_builder)
        mock_query_builder.should_receive('new_query').once().and_return(query)
        relation.get_parent().should_receive('fresh_timestamp').once().and_return(now)
        relation.should_receive('touch_if_touching').once()

        relation.attach(2, {'foo': 'bar'})

    def test_detach_remove_pivot_table_record(self):
        flexmock(BelongsToMany, touch_if_touching=lambda: True)
        relation = self._get_relation()
        query = flexmock()
        query.should_receive('from_').once().with_args('user_role').and_return(query)
        query.should_receive('where').once().with_args('user_id', 1).and_return(query)
        query.should_receive('where_in').once().with_args('role_id', [1, 2, 3])
        query.should_receive('delete').once().and_return(True)
        mock_query_builder = flexmock()
        relation.get_query().should_receive('get_query').and_return(mock_query_builder)
        mock_query_builder.should_receive('new_query').once().and_return(query)
        relation.should_receive('touch_if_touching').once()

        self.assertTrue(relation.detach([1, 2, 3]))

    def test_detach_with_single_id_remove_pivot_table_record(self):
        flexmock(BelongsToMany, touch_if_touching=lambda: True)
        relation = self._get_relation()
        query = flexmock()
        query.should_receive('from_').once().with_args('user_role').and_return(query)
        query.should_receive('where').once().with_args('user_id', 1).and_return(query)
        query.should_receive('where_in').once().with_args('role_id', [1])
        query.should_receive('delete').once().and_return(True)
        mock_query_builder = flexmock()
        relation.get_query().should_receive('get_query').and_return(mock_query_builder)
        mock_query_builder.should_receive('new_query').once().and_return(query)
        relation.should_receive('touch_if_touching').once()

        self.assertTrue(relation.detach(1))

    def test_detach_clears_all_records_when_no_ids(self):
        flexmock(BelongsToMany, touch_if_touching=lambda: True)
        relation = self._get_relation()
        query = flexmock()
        query.should_receive('from_').once().with_args('user_role').and_return(query)
        query.should_receive('where').once().with_args('user_id', 1).and_return(query)
        query.should_receive('where_in').never()
        query.should_receive('delete').once().and_return(True)
        mock_query_builder = flexmock()
        relation.get_query().should_receive('get_query').and_return(mock_query_builder)
        mock_query_builder.should_receive('new_query').once().and_return(query)
        relation.should_receive('touch_if_touching').once()

        self.assertTrue(relation.detach())

    def test_create_creates_new_model_and_insert_attachment_record(self):
        flexmock(BelongsToMany, attach=lambda: True)
        relation = self._get_relation()
        model = flexmock()
        relation.get_related().should_receive('new_instance').once().and_return(model).with_args({'foo': 'bar'})
        model.should_receive('save').once()
        model.should_receive('get_key').and_return('foo')
        relation.should_receive('attach').once().with_args('foo', {'bar': 'baz'}, True)

        self.assertEqual(model, relation.create({'foo': 'bar'}, {'bar': 'baz'}))

    def test_find_or_new_finds_model(self):
        flexmock(BelongsToMany)
        relation = self._get_relation()
        model = flexmock()
        model.foo = 'bar'
        relation.get_query().should_receive('find').with_args('foo', None).and_return(model)
        relation.get_related().should_receive('new_instance').never()

        self.assertEqual('bar', relation.find_or_new('foo').foo)

    def test_find_or_new_returns_new_model(self):
        flexmock(BelongsToMany)
        relation = self._get_relation()
        model = flexmock()
        model.foo = 'bar'
        relation.get_query().should_receive('find').with_args('foo', None).and_return(None)
        relation.get_related().should_receive('new_instance').once().and_return(model)

        self.assertEqual('bar', relation.find_or_new('foo').foo)

    def test_first_or_new_finds_first_model(self):
        flexmock(BelongsToMany)
        relation = self._get_relation()
        model = flexmock()
        model.foo = 'bar'
        relation.get_query().should_receive('where').with_args({'foo': 'bar'}).and_return(relation.get_query())
        relation.get_query().should_receive('first').once().and_return(model)
        relation.get_related().should_receive('new_instance').never()

        self.assertEqual('bar', relation.first_or_new({'foo': 'bar'}).foo)

    def test_first_or_new_returns_new_model(self):
        flexmock(BelongsToMany)
        relation = self._get_relation()
        model = flexmock()
        model.foo = 'bar'
        relation.get_query().should_receive('where').with_args({'foo': 'bar'}).and_return(relation.get_query())
        relation.get_query().should_receive('first').once().and_return(None)
        relation.get_related().should_receive('new_instance').once().and_return(model)

        self.assertEqual('bar', relation.first_or_new({'foo': 'bar'}).foo)

    def test_first_or_create_finds_first_model(self):
        flexmock(BelongsToMany)
        relation = self._get_relation()
        model = flexmock()
        model.foo = 'bar'
        relation.get_query().should_receive('where').with_args({'foo': 'bar'}).and_return(relation.get_query())
        relation.get_query().should_receive('first').once().and_return(model)
        relation.should_receive('create').never()

        self.assertEqual('bar', relation.first_or_create({'foo': 'bar'}).foo)

    def test_first_or_create_returns_new_model(self):
        flexmock(BelongsToMany)
        relation = self._get_relation()
        model = flexmock()
        model.foo = 'bar'
        relation.get_query().should_receive('where').with_args({'foo': 'bar'}).and_return(relation.get_query())
        relation.get_query().should_receive('first').once().and_return(None)
        relation.should_receive('create').once().with_args({'foo': 'bar'}, {}, True).and_return(model)

        self.assertEqual('bar', relation.first_or_create({'foo': 'bar'}).foo)

    def test_update_or_create_finds_first_mode_and_updates(self):
        flexmock(BelongsToMany)
        relation = self._get_relation()
        model = flexmock()
        model.foo = 'bar'
        relation.get_query().should_receive('where').with_args({'foo': 'bar'}).and_return(relation.get_query())
        relation.get_query().should_receive('first').once().and_return(model)
        model.should_receive('fill').once()
        model.should_receive('save').once()
        relation.should_receive('create').never()

        self.assertEqual('bar', relation.update_or_create({'foo': 'bar'}).foo)

    def test_update_or_create_returns_new_model(self):
        flexmock(BelongsToMany)
        relation = self._get_relation()
        model = flexmock()
        model.foo = 'bar'
        relation.get_query().should_receive('where').with_args({'foo': 'bar'}).and_return(relation.get_query())
        relation.get_query().should_receive('first').once().and_return(None)
        relation.should_receive('create').once().with_args({'bar': 'baz'}, None, True).and_return(model)

        self.assertEqual('bar', relation.update_or_create({'foo': 'bar'}, {'bar': 'baz'}).foo)

    def test_sync_syncs_intermediate_table_with_given_list(self):
        for list_ in [[2, 3, 4], ['2', '3', '4']]:
            flexmock(BelongsToMany)
            relation = self._get_relation()
            query = flexmock()
            query.should_receive('from_').once().with_args('user_role').and_return(query)
            query.should_receive('where').once().with_args('user_id', 1).and_return(query)
            mock_query_builder = flexmock()
            relation.get_query().should_receive('get_query').and_return(mock_query_builder)
            mock_query_builder.should_receive('new_query').once().and_return(query)
            query.should_receive('lists').once().with_args('role_id').and_return([1, list_[0], list_[1]])
            relation.should_receive('attach').once().with_args(list_[2], {}, False)
            relation.should_receive('detach').once().with_args([1])
            relation.get_related().should_receive('touches').and_return(False)
            relation.get_parent().should_receive('touches').and_return(False)

            self.assertEqual(
                {
                    'attached': [list_[2]],
                    'detached': [1],
                    'updated': []
                },
                relation.sync(list_)
            )

    def test_sync_syncs_intermediate_table_with_given_list_and_attributes(self):
        flexmock(BelongsToMany)
        relation = self._get_relation()
        query = flexmock()
        query.should_receive('from_').once().with_args('user_role').and_return(query)
        query.should_receive('where').once().with_args('user_id', 1).and_return(query)
        mock_query_builder = flexmock()
        relation.get_query().should_receive('get_query').and_return(mock_query_builder)
        mock_query_builder.should_receive('new_query').once().and_return(query)
        query.should_receive('lists').once().with_args('role_id').and_return([1, 2, 3])
        relation.should_receive('attach').once().with_args(4, {'foo': 'bar'}, False)
        relation.should_receive('update_existing_pivot').once().with_args(3, {'bar': 'baz'}, False).and_return(True)
        relation.should_receive('detach').once().with_args([1])
        relation.should_receive('touch_if_touching').once()
        relation.get_related().should_receive('touches').and_return(False)
        relation.get_parent().should_receive('touches').and_return(False)

        self.assertEqual(
            {
                'attached': [4],
                'detached': [1],
                'updated': [3]
            },
            relation.sync([2, {3: {'bar': 'baz'}}, {4: {'foo': 'bar'}}], )
        )

    def test_sync_does_not_return_values_that_were_not_updated(self):
        flexmock(BelongsToMany)
        relation = self._get_relation()
        query = flexmock()
        query.should_receive('from_').once().with_args('user_role').and_return(query)
        query.should_receive('where').once().with_args('user_id', 1).and_return(query)
        mock_query_builder = flexmock()
        relation.get_query().should_receive('get_query').and_return(mock_query_builder)
        mock_query_builder.should_receive('new_query').once().and_return(query)
        query.should_receive('lists').once().with_args('role_id').and_return([1, 2, 3])
        relation.should_receive('attach').once().with_args(4, {'foo': 'bar'}, False)
        relation.should_receive('update_existing_pivot').once().with_args(3, {'bar': 'baz'}, False).and_return(False)
        relation.should_receive('detach').once().with_args([1])
        relation.should_receive('touch_if_touching').once()
        relation.get_related().should_receive('touches').and_return(False)
        relation.get_parent().should_receive('touches').and_return(False)

        self.assertEqual(
            {
                'attached': [4],
                'detached': [1],
                'updated': []
            },
            relation.sync([2, {3: {'bar': 'baz'}}, {4: {'foo': 'bar'}}], )
        )

    def test_touch_method_syncs_timestamps(self):
        relation = self._get_relation()
        relation.get_related().should_receive('get_updated_at_column').and_return('updated_at')
        now = arrow.get().naive
        relation.get_related().should_receive('fresh_timestamp').and_return(now)
        relation.get_related().should_receive('get_qualified_key_name').and_return('table.id')
        relation.get_query().get_query().should_receive('select').once().with_args('table.id')\
            .and_return(relation.get_query().get_query())
        relation.get_query().should_receive('lists').once().and_return([1, 2, 3])
        query = flexmock()
        relation.get_related().should_receive('new_query').once().and_return(query)
        query.should_receive('where_in').once().with_args('id', [1, 2, 3]).and_return(query)
        query.should_receive('update').once().with_args({'updated_at': now})

        relation.touch()

    def test_touch_if_touching(self):
        flexmock(BelongsToMany)
        relation = self._get_relation()
        relation.should_receive('_touching_parent').once().and_return(True)
        relation.get_parent().should_receive('touch').once()
        relation.get_parent().should_receive('touches').once().with_args('relation_name').and_return(True)
        relation.should_receive('touch').once()

        relation.touch_if_touching()

    def test_sync_method_converts_collection_to_list_of_keys(self):
        flexmock(BelongsToMany)
        relation = self._get_relation()
        query = flexmock()
        query.should_receive('from_').once().with_args('user_role').and_return(query)
        query.should_receive('where').once().with_args('user_id', 1).and_return(query)
        mock_query_builder = flexmock()
        relation.get_query().should_receive('get_query').and_return(mock_query_builder)
        mock_query_builder.should_receive('new_query').once().and_return(query)
        query.should_receive('lists').once().with_args('role_id').and_return([1, 2, 3])

        collection = flexmock(Collection())
        collection.should_receive('model_keys').once().and_return([1, 2, 3])
        relation.should_receive('_format_sync_list').with_args([1, 2, 3]).and_return({1: {}, 2: {}, 3: {}})

        relation.sync(collection)

    def test_where_pivot_params_used_for_new_queries(self):
        flexmock(BelongsToMany)
        relation = self._get_relation()

        relation.get_query().should_receive('where').once().and_return(relation)

        query = flexmock()
        mock_query_builder = flexmock()
        relation.get_query().should_receive('get_query').and_return(mock_query_builder)
        mock_query_builder.should_receive('new_query').once().and_return(query)

        query.should_receive('from_').once().with_args('user_role').and_return(query)

        query.should_receive('where').once().with_args('user_id', 1).and_return(query)

        query.should_receive('where').once().with_args('foo', '=', 'bar', 'and').and_return(query)

        query.should_receive('lists').once().with_args('role_id').and_return([1, 2, 3])
        relation.should_receive('_format_sync_list').with_args([1, 2, 3]).and_return({1: {}, 2: {}, 3: {}})

        relation = relation.where_pivot('foo', '=', 'bar')
        relation.sync([1, 2, 3])

    def _get_relation(self):
        builder, parent = self._get_relation_arguments()[:2]

        return BelongsToMany(builder, parent, 'user_role', 'user_id', 'role_id', 'relation_name')

    def _get_relation_arguments(self):
        parent = flexmock(Model())
        parent.should_receive('get_key').and_return(1)
        parent.should_receive('get_created_at_column').and_return('created_at')
        parent.should_receive('get_updated_at_column').and_return('updated_at')

        query = flexmock(QueryBuilder(MockConnection().prepare_mock(), QueryGrammar(), QueryProcessor()))
        flexmock(Builder)
        builder = Builder(query)
        builder.should_receive('get_query').and_return(query)
        related = flexmock(Model())
        builder.set_model(related)
        builder.should_receive('get_model').and_return(related)

        related.should_receive('get_key_name').and_return('id')
        related.should_receive('get_table').and_return('roles')
        related.should_receive('new_pivot').replace_with(lambda *args: Pivot(*args))

        builder.get_query().should_receive('join').once().with_args('user_role', 'roles.id', '=', 'user_role.role_id')
        builder.should_receive('where').once().with_args('user_role.user_id', '=', 1)

        return builder, parent, 'user_role', 'user_id', 'role_id', 'relation_id'


class OrmBelongsToManyModelStub(Model):

    __guarded__ = []


class OrmBelongsToManyModelPivotStub(Model):

    __guarded__ = []

    def __init__(self):
        super(OrmBelongsToManyModelPivotStub, self).__init__()

        self.pivot = OrmBelongsToManyPivotStub()


class OrmBelongsToManyPivotStub(object):
    pass



# -*- coding: utf-8 -*-


import arrow
from flexmock import flexmock, flexmock_teardown
from ... import EloquentTestCase

from eloquent.query.builder import QueryBuilder
from eloquent.query.grammars import QueryGrammar
from eloquent.query.expression import QueryExpression
from eloquent.orm.builder import Builder
from eloquent.orm.model import Model
from eloquent.orm.relations import HasOne
from eloquent.orm.collection import Collection


class OrmHasOneTestCase(EloquentTestCase):

    def tearDown(self):
        flexmock_teardown()

    def test_save_method_set_foreign_key_on_model(self):
        relation = self._get_relation()
        mock_model = flexmock(Model(), save=lambda: True)
        mock_model.should_receive('save').once().and_return(True)
        result = relation.save(mock_model)

        attributes = result.get_attributes()
        self.assertEqual(1, attributes['foreign_key'])

    def test_create_properly_creates_new_model(self):
        relation = self._get_relation()
        created = flexmock(Model(), save=lambda: True, set_attribute=lambda: None)
        created.should_receive('save').once().and_return(True)
        relation.get_related().should_receive('new_instance').once().with_args({'name': 'john'}).and_return(created)
        created.should_receive('set_attribute').with_args('foreign_key', 1)

        self.assertEqual(created, relation.create(name='john'))

    def test_update_updates_models_with_timestamps(self):
        relation = self._get_relation()
        relation.get_related().should_receive('uses_timestamps').once().and_return(True)
        now = arrow.get()
        relation.get_related().should_receive('fresh_timestamp').once().and_return(now)
        relation.get_query().should_receive('update').once().with_args({'foo': 'bar', 'updated_at': now}).and_return('results')

        self.assertEqual('results', relation.update(foo='bar'))

    def test_relation_is_properly_initialized(self):
        relation = self._get_relation()
        model = flexmock(Model())
        model.should_receive('set_relation').once().with_args('foo', None)
        models = relation.init_relation([model], 'foo')

        self.assertEqual([model], models)

    def test_eager_constraints_are_properly_added(self):
        relation = self._get_relation()
        relation.get_query().get_query().should_receive('where_in').once().with_args('table.foreign_key', [1, 2])

        model1 = OrmHasOneModelStub()
        model1.id = 1
        model2 = OrmHasOneModelStub()
        model2.id = 2

        relation.add_eager_constraints([model1, model2])

    def test_models_are_properly_matched_to_parents(self):
        relation = self._get_relation()

        result1 = OrmHasOneModelStub()
        result1.foreign_key = 1
        result2 = OrmHasOneModelStub()
        result2.foreign_key = 2

        model1 = OrmHasOneModelStub()
        model1.id = 1
        model2 = OrmHasOneModelStub()
        model2.id = 2
        model3 = OrmHasOneModelStub()
        model3.id = 3

        models = relation.match([model1, model2, model3], Collection([result1, result2]), 'foo')

        self.assertEqual(1, models[0].foo.foreign_key)
        self.assertEqual(2, models[1].foo.foreign_key)
        self.assertFalse(hasattr(models[2], 'foo'))

    def test_relation_count_query_can_be_built(self):
        relation = self._get_relation()
        query = flexmock(QueryBuilder(None, QueryGrammar(), None))
        builder = Builder(query)
        builder.get_query().should_receive('select').once()
        relation.get_parent().should_receive('get_table').and_return('table')
        builder.should_receive('where').once().with_args('table.foreign_key', '=', QueryExpression)
        parent_query = flexmock(QueryBuilder(None, None, None))
        relation.get_query().should_receive('get_query').and_return(parent_query)
        grammar = flexmock()
        parent_query.should_receive('get_grammar').once().and_return(grammar)
        grammar.should_receive('wrap').once().with_args('table.id')

        relation.get_relation_count_query(builder, builder)

    def _get_relation(self):
        flexmock(Builder)
        query = flexmock(QueryBuilder(None, QueryGrammar(), None))
        builder = Builder(query)
        builder.should_receive('where').with_args('table.foreign_key', '=', 1)
        related = flexmock(Model())
        builder.should_receive('get_model').and_return(related)
        parent = flexmock(Model())
        parent.should_receive('get_attribute').with_args('id').and_return(1)
        parent.should_receive('get_created_at_column').and_return('created_at')
        parent.should_receive('get_updated_at_column').and_return('updated_at')
        parent.should_receive('new_query').and_return(builder)

        return HasOne(builder, parent, 'table.foreign_key', 'id')


class OrmHasOneModelStub(Model):

    pass



# -*- coding: utf-8 -*-


import arrow
from flexmock import flexmock, flexmock_teardown
from ... import EloquentTestCase, mock

from eloquent.query.builder import QueryBuilder
from eloquent.query.grammars import QueryGrammar
from eloquent.query.expression import QueryExpression
from eloquent.orm.builder import Builder
from eloquent.orm.model import Model
from eloquent.orm.relations import BelongsTo
from eloquent.orm.collection import Collection


class OrmBelongsToTestCase(EloquentTestCase):

    def tearDown(self):
        flexmock_teardown()

    def test_update_retrieve_model_and_updates(self):
        relation = self._get_relation()
        mock = flexmock(Model())
        mock.should_receive('fill').once().with_args({'foo': 'bar'}).and_return(mock)
        mock.should_receive('save').once().and_return(True)
        relation.get_query().should_receive('first').once().and_return(mock)

        self.assertTrue(relation.update({'foo': 'bar'}))

    def test_relation_is_properly_initialized(self):
        relation = self._get_relation()
        model = flexmock(Model())
        model.should_receive('set_relation').once().with_args('foo', None)
        models = relation.init_relation([model], 'foo')

        self.assertEqual([model], models)

    def test_eager_constraints_are_properly_added(self):
        relation = self._get_relation()
        relation.get_query().get_query().should_receive('where_in').once()\
            .with_args('relation.id', ['foreign.value', 'foreign.value.two'])

        model1 = OrmBelongsToModelStub()
        model2 = OrmBelongsToModelStub()
        model2.foreign_key = 'foreign.value'
        model3 = AnotherOrmBelongsToModelStub()
        model3.foreign_key = 'foreign.value.two'
        models = [model1, model2, model3]

        relation.add_eager_constraints(models)

    def test_models_are_properly_matched_to_parents(self):
        relation = self._get_relation()

        result1 = flexmock()
        result1.should_receive('get_attribute').with_args('id').and_return(1)
        result2 = flexmock()
        result2.should_receive('get_attribute').with_args('id').and_return(2)

        model1 = OrmBelongsToModelStub()
        model1.foreign_key = 1
        model2 = OrmBelongsToModelStub()
        model2.foreign_key = 2

        models = relation.match([model1, model2], Collection([result1, result2]), 'foo')

        self.assertEqual(1, models[0].foo.get_attribute('id'))
        self.assertEqual(2, models[1].foo.get_attribute('id'))

    def test_associate_sets_foreign_key_on_model(self):
        parent = Model()
        parent.foreign_key = 'foreign.value'
        parent.get_attribute = mock.MagicMock(return_value='foreign.value')
        parent.set_attribute = mock.MagicMock()
        parent.set_relation = mock.MagicMock()
        relation = self._get_relation(parent)
        associate = flexmock(Model())
        associate.should_receive('get_attribute').once().with_args('id').and_return(1)

        relation.associate(associate)

        parent.get_attribute.assert_has_calls([
            mock.call('foreign_key'),
            mock.call('foreign_key')
        ])
        parent.set_attribute.assert_has_calls([
            mock.call('foreign_key', 1)
        ])
        parent.set_relation.assert_called_once_with('relation', associate)

    def _get_relation(self, parent=None):
        flexmock(Builder)
        query = flexmock(QueryBuilder(None, QueryGrammar(), None))
        builder = Builder(query)
        builder.should_receive('where').with_args('relation.id', '=', 'foreign.value')
        related = flexmock(Model())
        related.should_receive('get_key_name').and_return('id')
        related.should_receive('get_table').and_return('relation')
        builder.should_receive('get_model').and_return(related)
        if parent is None:
            parent = OrmBelongsToModelStub()
        parent.foreign_key = 'foreign.value'

        return BelongsTo(builder, parent, 'foreign_key', 'id', 'relation')


class OrmBelongsToModelStub(Model):

    foreign_key = 'foreign.value'


class AnotherOrmBelongsToModelStub(Model):

    foreign_key = 'foreign.value.two'



# -*- coding: utf-8 -*-


import arrow
from flexmock import flexmock, flexmock_teardown
from ... import EloquentTestCase
from ...utils import MockConnection

from eloquent.query.builder import QueryBuilder
from eloquent.query.grammars import QueryGrammar
from eloquent.query.processors import QueryProcessor
from eloquent.query.expression import QueryExpression
from eloquent.orm.builder import Builder
from eloquent.orm.model import Model
from eloquent.orm.relations import MorphToMany
from eloquent.orm.relations.pivot import Pivot
from eloquent.orm.collection import Collection


class OrmMorphToManyTestCase(EloquentTestCase):

    def tearDown(self):
        flexmock_teardown()

    def test_eager_constraints_are_properly_added(self):
        relation = self._get_relation()
        relation.get_query().get_query().should_receive('where_in').once().with_args('taggables.taggable_id', [1, 2])
        relation.get_query().should_receive('where').once()\
            .with_args('taggables.taggable_type', relation.get_parent().__class__.__name__)
        model1 = OrmMorphToManyModelStub()
        model1.id = 1
        model2 = OrmMorphToManyModelStub()
        model2.id = 2

        relation.add_eager_constraints([model1, model2])

    def test_attach_inserts_pivot_table_record(self):
        flexmock(MorphToMany, touch_if_touching=lambda: True)
        relation = self._get_relation()
        query = flexmock()
        query.should_receive('from_').once().with_args('taggables').and_return(query)
        query.should_receive('insert').once()\
            .with_args(
                [{
                    'taggable_id': 1,
                    'taggable_type': relation.get_parent().__class__.__name__,
                    'tag_id': 2,
                    'foo': 'bar',
                }])\
            .and_return(True)
        mock_query_builder = flexmock()
        relation.get_query().should_receive('get_query').and_return(mock_query_builder)
        mock_query_builder.should_receive('new_query').once().and_return(query)
        relation.should_receive('touch_if_touching').once()

        relation.attach(2, {'foo': 'bar'})

    def test_detach_remove_pivot_table_record(self):
        flexmock(MorphToMany, touch_if_touching=lambda: True)
        relation = self._get_relation()
        query = flexmock()
        query.should_receive('from_').once().with_args('taggables').and_return(query)
        query.should_receive('where').once().with_args('taggable_id', 1).and_return(query)
        query.should_receive('where').once()\
            .with_args('taggable_type', relation.get_parent().__class__.__name__).and_return(query)
        query.should_receive('where_in').once().with_args('tag_id', [1, 2, 3])
        query.should_receive('delete').once().and_return(True)
        mock_query_builder = flexmock()
        relation.get_query().should_receive('get_query').and_return(mock_query_builder)
        mock_query_builder.should_receive('new_query').once().and_return(query)
        relation.should_receive('touch_if_touching').once()

        self.assertTrue(relation.detach([1, 2, 3]))

    def test_detach_clears_all_records_when_no_ids(self):
        flexmock(MorphToMany, touch_if_touching=lambda: True)
        relation = self._get_relation()
        query = flexmock()
        query.should_receive('from_').once().with_args('taggables').and_return(query)
        query.should_receive('where').once().with_args('taggable_id', 1).and_return(query)
        query.should_receive('where').once()\
            .with_args('taggable_type', relation.get_parent().__class__.__name__).and_return(query)
        query.should_receive('where_in').never()
        query.should_receive('delete').once().and_return(True)
        mock_query_builder = flexmock()
        relation.get_query().should_receive('get_query').and_return(mock_query_builder)
        mock_query_builder.should_receive('new_query').once().and_return(query)
        relation.should_receive('touch_if_touching').once()

        self.assertTrue(relation.detach())

    def _get_relation(self):
        builder, parent = self._get_relation_arguments()[:2]

        return MorphToMany(builder, parent, 'taggable', 'taggables', 'taggable_id', 'tag_id')

    def _get_relation_arguments(self):
        parent = flexmock(Model())
        parent.should_receive('get_morph_class').and_return(parent.__class__.__name__)
        parent.should_receive('get_key').and_return(1)
        parent.should_receive('get_created_at_column').and_return('created_at')
        parent.should_receive('get_updated_at_column').and_return('updated_at')

        query = flexmock(QueryBuilder(MockConnection().prepare_mock(), QueryGrammar(), QueryProcessor()))
        flexmock(Builder)
        builder = Builder(query)
        builder.should_receive('get_query').and_return(query)
        related = flexmock(Model())
        builder.set_model(related)
        builder.should_receive('get_model').and_return(related)

        related.should_receive('get_key_name').and_return('id')
        related.should_receive('get_table').and_return('tags')
        related.should_receive('get_morph_class').and_return(parent.__class__.__name__)

        builder.get_query().should_receive('join').once().with_args('taggables', 'tags.id', '=', 'taggables.tag_id')
        builder.should_receive('where').once().with_args('taggables.taggable_id', '=', 1)
        builder.should_receive('where').once().with_args('taggables.taggable_type', parent.__class__.__name__)

        return builder, parent, 'taggable', 'taggables', 'taggable_id', 'tag_id', 'relation_name', False


class OrmMorphToManyModelStub(Model):

    __guarded__ = []


class OrmMorphToManyModelPivotStub(Model):

    __guarded__ = []

    def __init__(self):
        super(OrmMorphToManyModelPivotStub, self).__init__()

        self.pivot = OrmMorphToManyPivotStub()


class OrmMorphToManyPivotStub(object):
    pass



# -*- coding: utf-8 -*-

from flexmock import flexmock, flexmock_teardown
from .. import EloquentTestCase, mock
from ..utils import MockModel, MockQueryBuilder, MockConnection, MockProcessor

from eloquent.query.grammars.grammar import QueryGrammar
from eloquent.query.builder import QueryBuilder
from eloquent.orm.builder import Builder
from eloquent.orm.model import Model
from eloquent.exceptions.orm import ModelNotFound
from eloquent.orm.collection import Collection


class BuilderTestCase(EloquentTestCase):

    def tearDown(self):
        flexmock_teardown()

    def test_find_method(self):
        builder = Builder(self.get_mock_query_builder())
        builder.set_model(self.get_mock_model())
        builder.get_query().where = mock.MagicMock()
        builder.first = mock.MagicMock(return_value='baz')

        result = builder.find('bar', ['column'])

        builder.get_query().where.assert_called_once_with(
            'foo_table.foo', '=', 'bar'
        )
        self.assertEqual('baz', result)

    def test_find_or_new_model_found(self):
        model = self.get_mock_model()
        model.find_or_new = mock.MagicMock(return_value='baz')

        builder = Builder(self.get_mock_query_builder())
        builder.set_model(model)
        builder.get_query().where = mock.MagicMock()
        builder.first = mock.MagicMock(return_value='baz')

        expected = model.find_or_new('bar', ['column'])
        result = builder.find('bar', ['column'])

        builder.get_query().where.assert_called_once_with(
            'foo_table.foo', '=', 'bar'
        )
        self.assertEqual(expected, result)

    def test_find_or_new_model_not_found(self):
        model = self.get_mock_model()
        model.find_or_new = mock.MagicMock(return_value=self.get_mock_model())

        builder = Builder(self.get_mock_query_builder())
        builder.set_model(model)
        builder.get_query().where = mock.MagicMock()
        builder.first = mock.MagicMock(return_value=None)

        result = model.find_or_new('bar', ['column'])
        find_result = builder.find('bar', ['column'])

        builder.get_query().where.assert_called_once_with(
            'foo_table.foo', '=', 'bar'
        )
        self.assertIsNone(find_result)
        self.assertIsInstance(result, Model)

    def test_find_or_fail_raises_model_not_found_exception(self):
        model = self.get_mock_model()

        builder = Builder(self.get_mock_query_builder())
        builder.set_model(model)
        builder.get_query().where = mock.MagicMock()
        builder.first = mock.MagicMock(return_value=None)

        self.assertRaises(
            ModelNotFound,
            builder.find_or_fail,
            'bar',
            ['column']
        )

        builder.get_query().where.assert_called_once_with(
            'foo_table.foo', '=', 'bar'
        )

        builder.first.assert_called_once_with(
            ['column']
        )

    def test_find_or_fail_with_many_raises_model_not_found_exception(self):
        model = self.get_mock_model()

        builder = Builder(self.get_mock_query_builder())
        builder.set_model(model)
        builder.get_query().where_in = mock.MagicMock()
        builder.get = mock.MagicMock(return_value=Collection([1]))

        self.assertRaises(
            ModelNotFound,
            builder.find_or_fail,
            [1, 2],
            ['column']
        )

        builder.get_query().where_in.assert_called_once_with(
            'foo_table.foo', [1, 2]
        )

        builder.get.assert_called_once_with(
            ['column']
        )

    def test_first_or_fail_raises_model_not_found_exception(self):
        model = self.get_mock_model()

        builder = Builder(self.get_mock_query_builder())
        builder.set_model(model)
        builder.first = mock.MagicMock(return_value=None)

        self.assertRaises(
            ModelNotFound,
            builder.first_or_fail,
            ['column']
        )

        builder.first.assert_called_once_with(
            ['column']
        )

    def test_find_with_many(self):
        model = self.get_mock_model()

        builder = Builder(self.get_mock_query_builder())
        builder.set_model(model)
        builder.get_query().where_in = mock.MagicMock()
        builder.get = mock.MagicMock(return_value='baz')

        result = builder.find([1, 2], ['column'])
        self.assertEqual('baz', result)

        builder.get_query().where_in.assert_called_once_with(
            'foo_table.foo', [1, 2]
        )

        builder.get.assert_called_once_with(
            ['column']
        )

    def test_first(self):
        model = self.get_mock_model()

        builder = Builder(self.get_mock_query_builder())
        builder.set_model(model)
        builder.take = mock.MagicMock(return_value=builder)
        builder.get = mock.MagicMock(return_value=Collection(['bar']))

        result = builder.first()
        self.assertEqual('bar', result)

        builder.take.assert_called_once_with(
            1
        )

        builder.get.assert_called_once_with(
            ['*']
        )

    def test_get_loads_models_and_hydrates_eager_relations(self):
        flexmock(Builder)
        builder = Builder(self.get_mock_query_builder())
        builder.should_receive('get_models').with_args(['foo']).and_return(['bar'])
        builder.should_receive('eager_load_relations').with_args(['bar']).and_return(['bar', 'baz'])
        builder.set_model(self.get_mock_model())
        builder.get_model().new_collection = mock.MagicMock(return_value=Collection(['bar', 'baz']))

        results = builder.get(['foo'])
        self.assertEqual(['bar', 'baz'], results.all())

        builder.get_model().new_collection.assert_called_with(['bar', 'baz'])

    def test_get_does_not_eager_relations_when_no_results_are_returned(self):
        flexmock(Builder)
        builder = Builder(self.get_mock_query_builder())
        builder.should_receive('get_models').with_args(['foo']).and_return(['bar'])
        builder.should_receive('eager_load_relations').with_args(['bar']).and_return([])
        builder.set_model(self.get_mock_model())
        builder.get_model().new_collection = mock.MagicMock(return_value=Collection([]))

        results = builder.get(['foo'])
        self.assertEqual([], results.all())

        builder.get_model().new_collection.assert_called_with([])

    def test_pluck_with_model_found(self):
        builder = Builder(self.get_mock_query_builder())

        model = {'name': 'foo'}
        builder.first = mock.MagicMock(return_value=model)

        self.assertEqual('foo', builder.pluck('name'))

        builder.first.assert_called_once_with(
            ['name']
        )

    def test_pluck_with_model_not_found(self):
        builder = Builder(self.get_mock_query_builder())

        builder.first = mock.MagicMock(return_value=None)

        self.assertIsNone(builder.pluck('name'))

    def test_chunk(self):
        builder = Builder(self.get_mock_query_builder())
        results = [['foo1', 'foo2'], ['foo3'], []]
        builder.for_page = mock.MagicMock(return_value=builder)
        builder.get = mock.MagicMock(side_effect=results)

        i = 0
        for result in builder.chunk(2):
            self.assertEqual(result, results[i])

            i += 1

        builder.for_page.assert_has_calls([
            mock.call(1, 2),
            mock.call(2, 2),
            mock.call(3, 2)
        ])

    # TODO: lists with get mutators

    def test_lists_without_model_getters(self):
        builder = self.get_builder()
        builder.get_query().get = mock.MagicMock(return_value=[{'name': 'bar'}, {'name': 'baz'}])
        builder.set_model(self.get_mock_model())
        builder.get_model().has_get_mutator = mock.MagicMock(return_value=False)

        result = builder.lists('name')
        self.assertEqual(['bar', 'baz'], result)

        builder.get_query().get.assert_called_once_with(['name'])

    def test_get_models_hydrates_models(self):
        builder = Builder(self.get_mock_query_builder())
        records = [{
            'name': 'john', 'age': 26
        }, {
            'name': 'jane', 'age': 28
        }]

        builder.get_query().get = mock.MagicMock(return_value=records)
        model = self.get_mock_model()
        builder.set_model(model)
        model.get_connection_name = mock.MagicMock(return_value='foo_connection')
        model.hydrate = mock.MagicMock(return_value=Collection(['hydrated']))
        models = builder.get_models(['foo'])

        self.assertEqual(models, ['hydrated'])

        model.get_table.assert_called_once_with()
        model.get_connection_name.assert_called_once_with()
        model.hydrate.assert_called_once_with(
            records, 'foo_connection'
        )

    def test_eager_load_relations_load_top_level_relationships(self):
        flexmock(Builder)
        builder = Builder(flexmock(QueryBuilder(None, None, None)))
        nop1 = lambda: None
        nop2 = lambda: None
        builder.set_eager_loads({'foo': nop1, 'foo.bar': nop2})
        builder.should_receive('_load_relation').with_args(['models'], 'foo', nop1).and_return(['foo'])

        results = builder.eager_load_relations(['models'])
        self.assertEqual(['foo'], results)

    def test_eager_load_accept_queries(self):
        model = OrmBuilderTestModelCloseRelated()
        flexmock(Builder)
        builder = Builder(flexmock(QueryBuilder(None, None, None)))
        nop1 = OrmBuilderTestModelFarRelatedStub.where('id', 5)
        builder.set_eager_loads({'foo': nop1})
        relation = flexmock()
        relation.should_receive('add_eager_constraints').once().with_args(['models'])
        relation.should_receive('init_relation').once().with_args(['models'], 'foo').and_return(['models'])
        relation.should_receive('get_eager').once().and_return(['results'])
        relation.should_receive('match').once()\
            .with_args(['models'], ['results'], 'foo').and_return(['foo'])
        builder.should_receive('get_relation').once().with_args('foo').and_return(relation)
        relation.should_receive('merge_query').with_args(nop1).and_return(relation)

        results = builder.eager_load_relations(['models'])
        self.assertEqual(['foo'], results)

    def test_relationship_eager_load_process(self):
        proof = flexmock()
        flexmock(Builder)
        builder = Builder(flexmock(QueryBuilder(None, None, None)))

        def callback(q):
            proof.foo = q

        builder.set_eager_loads({'orders': callback})
        relation = flexmock()
        relation.should_receive('add_eager_constraints').once().with_args(['models'])
        relation.should_receive('init_relation').once().with_args(['models'], 'orders').and_return(['models'])
        relation.should_receive('get_eager').once().and_return(['results'])
        relation.should_receive('match').once()\
            .with_args(['models'], ['results'], 'orders').and_return(['models.matched'])
        builder.should_receive('get_relation').once().with_args('orders').and_return(relation)
        results = builder.eager_load_relations(['models'])

        self.assertEqual(['models.matched'], results)
        self.assertEqual(relation, proof.foo)

    def test_get_relation_properly_sets_nested_relationships(self):
        flexmock(Builder)
        builder = Builder(flexmock(QueryBuilder(None, None, None)))
        model = flexmock(Model())
        relation = flexmock()
        model.set_relation('orders', relation)
        builder.set_model(model)
        relation_query = flexmock()
        relation.should_receive('get_query').and_return(relation_query)
        relation_query.should_receive('with_').once().with_args({'lines': None, 'lines.details': None})
        builder.set_eager_loads({
            'orders': None,
            'orders.lines': None,
            'orders.lines.details': None
        })

        relation = builder.get_relation('orders')

    def test_query_passthru(self):
        builder = self.get_builder()
        builder.get_query().foobar = mock.MagicMock(return_value='foo')

        self.assertIsInstance(builder.foobar(), Builder)
        self.assertEqual(builder.foobar(), builder)

        builder = self.get_builder()
        builder.get_query().insert = mock.MagicMock(return_value='foo')

        self.assertEqual('foo', builder.insert(['bar']))

        builder.get_query().insert.assert_called_once_with(['bar'])

    # TODO: test query scopes

    def test_simple_where(self):
        builder = self.get_builder()
        builder.get_query().where = mock.MagicMock()
        result = builder.where('foo', '=', 'bar')

        self.assertEqual(builder, result)

        builder.get_query().where.assert_called_once_with('foo', '=', 'bar', 'and')

    def test_nested_where(self):
        nested_query = self.get_builder()
        nested_raw_query = self.get_mock_query_builder()
        nested_query.get_query = mock.MagicMock(return_value=nested_raw_query)
        model = self.get_mock_model()
        builder = self.get_builder()
        builder.set_model(model)
        builder.get_query().add_nested_where_query = mock.MagicMock()

        result = builder.where(nested_query)
        self.assertEqual(builder, result)

        builder.get_query().add_nested_where_query.assert_called_once_with(nested_raw_query, 'and')

    # TODO: nested query with scopes

    def test_delete_override(self):
        builder = self.get_builder()

        builder.on_delete(lambda builder_: {'foo': builder_})

        self.assertEqual({'foo': builder}, builder.delete())

    def test_has_nested(self):
        builder = OrmBuilderTestModelParentStub.where_has('foo', lambda q: q.has('bar'))

        result = OrmBuilderTestModelParentStub.has('foo.bar').to_sql()

        self.assertEqual(builder.to_sql(), result)

    def test_has_nested_with_constraints(self):
        model = OrmBuilderTestModelParentStub

        builder = model.where_has('foo', lambda q: q.where_has('bar', lambda q: q.where('baz', 'bam'))).to_sql()

        result = model.where_has('foo.bar', lambda q: q.where('baz', 'bam')).to_sql()

        self.assertEqual(builder, result)

    def get_builder(self):
        return Builder(self.get_mock_query_builder())

    def get_mock_model(self):
        model = MockModel().prepare_mock()

        return model

    def get_mock_query_builder(self):
        connection = MockConnection().prepare_mock()
        processor = MockProcessor().prepare_mock()

        builder = MockQueryBuilder(
            connection,
            QueryGrammar(),
            processor
        ).prepare_mock()

        return builder


class OrmBuilderTestModelFarRelatedStub(Model):

    pass


class OrmBuilderTestModelCloseRelated(Model):

    @property
    def bar(self):
        return self.has_many(OrmBuilderTestModelFarRelatedStub)


class OrmBuilderTestModelParentStub(Model):

    @property
    def foo(self):
        return self.belongs_to(OrmBuilderTestModelCloseRelated)
